!> @file libmtime.f90
!!
!! @brief Providing the Fortran language bindings for libmtime
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!! @defgroup FortranBindings libmtime Fortran language bindings
!! @{
!!
!___________________________________________________________________________________________________________
!>
!! @brief Provides the calendar to the users, abstracting the different calendars available.
!!
!! @details
!!
!! Three calendar types are provided:
!!
!!   - a proleptic Gregorian calendar
!!   - a calendar with 365 days per year without leap years
!!   - a calendar with 360 days per year and each month having 30 days
!!
!! To use a specific calendar a call to setCalendar with the respective selector must be
!! done. The implementation is based on a singleton concept meaning that only one calendar
!! can be active at a time. To release a calendar a call to resetCalendar has to be done.
!!
!___________________________________________________________________________________________________________
module mtime_calendar
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_char, c_null_char, c_ptr, c_associated, c_int64_t
  !
  implicit none
  !
  private
  !   
  enum, bind(c)
    enumerator :: calendar_not_set    = 0   ! calendar is not defined yet
    enumerator :: proleptic_gregorian = 1   ! proleptic Gregorian calendar
    enumerator :: year_of_365_days    = 2   ! 365 day year without leap years
    enumerator :: year_of_360_days    = 3   ! 360 day year with 30 day months
  end enum
  !
  public :: max_calendar_str_len
  public :: calendar_not_set, proleptic_gregorian, year_of_365_days, year_of_360_days
  public :: setCalendar
  public :: resetCalendar
  public :: calendarType
  public :: calendarToString
  !
  !> provides a string length for toString 
  integer, parameter :: max_calendar_str_len = 32
  !
  !> @cond DOXYGEN_IGNORE_THIS
  interface
    !
    subroutine setCalendar(ct) bind(c, name='initCalendar') !TESTED-OK
      import :: c_int
      integer(c_int), value :: ct
    end subroutine setCalendar
    !
    subroutine resetCalendar() bind(c, name='freeCalendar') !TESTED-OK
    end subroutine resetCalendar
    !
    function calendarType() bind(c, name='getCalendarType') !TESTED-OK
      import :: c_int
      integer(c_int) :: calendarType
    end function calendarType
    !
    function my_calendartostring(calendar) result(c_pointer) bind(c, name='calendarToString') !TESTED-OK
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: calendar
    end function my_calendartostring
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !
#ifdef DOXYGEN_DOCUMENTATION_ONLY
  !> 
  !! @brief Initialize a new calendar.
  !!
  !! setCalendar is done at the very begining to select one of the
  !! provided calendar libraries. It intializes the calendar to one of:
  !!
  !! - proleptic_gregorian
  !! - year_of_365_days
  !! - year_of_360_days
  !! 
  !! The calendar type and hence it's behaviour (Calendar to Julian
  !! conversion and vice versa) is fixed for the lifetime of the
  !! selected calendar.  Attempts to change the calendar type on the
  !! fly is discouraged. The lib has built-in checks to reject
  !! change attempts at run time.  However, a calendar can be
  !! "re-initialized" after calling resetCalendar(), but this is not
  !! advised. 
  !!
  !! MANTRA: Know what you are doing before you do it and do it
  !! right the first time.
  !!
  !! @param[in]  ct      the calendar type
  subroutine setCalendar(ct) bind(c, name='initCalendar') !TESTED-OK
    integer(c_int), value :: ct
  end subroutine setCalendar
  !>
  !! @brief called to discard the selected calendar type 
  subroutine resetCalendar() bind(c, name='freeCalendar') !TESTED-OK
  end subroutine resetCalendar
  !>
  !! @brief to get an idea, which calendar type is selected
  !!
  !! @return an integer denoting the calendar in use
  function calendarType() bind(c, name='getCalendarType') !TESTED-OK
    integer(c_int) :: calendarType
  end function calendarType
  !
#endif
  !>
  !! @brief convert the calendar identifier into a human readable string
  !!
  !! @param[out]       string      the calendar type verbose
  !! @param[out]       errno       optional, error message 
  !!
  subroutine calendarToString(string, errno) !TESTED-OK
    character(len=max_calendar_str_len), intent(out) :: string
    integer :: i
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_calendartostring(string)
    if (.not. c_associated(c_pointer)) then
      if (present(errno)) errno = 0 * 100 + 4
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine calendarToString
  !
end module mtime_calendar


MODULE mtime_juliandelta
  !
  USE, INTRINSIC :: iso_c_binding, ONLY: c_int64_t, c_char, c_ptr, c_loc, &
       c_associated, c_f_pointer
  !
  IMPLICIT NONE
  !
  PRIVATE
  !
  PUBLIC :: juliandelta
  PUBLIC :: newJulianDelta
  PUBLIC :: deallocateJulianDelta
  !
  TYPE, bind(c) :: juliandelta
    CHARACTER(c_char)  :: sign
    INTEGER(c_int64_t) :: day 
    INTEGER(c_int64_t) :: ms  
  END TYPE juliandelta

  INTERFACE
    FUNCTION my_newjuliandelta(sign, day, ms) RESULT(c_pointer) bind(c, name='newJulianDelta')
      import :: c_int64_t, c_char, c_ptr
      TYPE(c_ptr)               :: c_pointer
      CHARACTER(c_char)         :: sign
      INTEGER(c_int64_t), value :: day
      INTEGER(c_int64_t), value :: ms
    end function my_newjuliandelta
    !
    subroutine my_deallocatejuliandelta(jd) bind(c,name='deallocateJulianDelta')
      import :: c_ptr
      type(c_ptr), value :: jd
    end subroutine my_deallocatejuliandelta
  END INTERFACE

CONTAINS

  FUNCTION newJuliandelta(sign,day, ms, errno) RESULT(ret_juliandelta) !OK-TESTED.
    TYPE(juliandelta), POINTER :: ret_juliandelta
    CHARACTER(c_char),  INTENT(in) :: sign
    INTEGER(c_int64_t), INTENT(in) :: day
    INTEGER(c_int64_t), INTENT(in) :: ms
    type(c_ptr)                    :: c_pointer
    integer, optional              :: errno

    IF (PRESENT(errno)) errno = 0
    c_pointer = my_newjuliandelta(sign,day, ms)
    IF ((.NOT. c_associated(c_pointer)) .AND. PRESENT(errno)) errno = 1 * 100 + 1
    CALL c_f_pointer(c_pointer, ret_juliandelta)
  END FUNCTION newJuliandelta
  !
  !>
  !! @brief destructor for a Julian delta
  SUBROUTINE deallocateJuliandelta(my_juliandelta) !OK-TESTED.
    TYPE(juliandelta), POINTER :: my_juliandelta
    CALL my_deallocatejuliandelta(c_loc(my_juliandelta))
    my_juliandelta => NULL()
  END SUBROUTINE deallocateJuliandelta

END MODULE mtime_juliandelta


!>
!! @brief Julian Day Calendar and some operations supported on julian dates.
!!
!! @details
!___________________________________________________________________________________________________________
module mtime_julianday
  !
  use, intrinsic :: iso_c_binding, only: c_int64_t, c_char, c_null_char, c_ptr, c_loc, &
                                       & c_f_pointer, c_associated
  !
  implicit none
  !
  private
  !
  public :: max_julianday_str_len
  public :: julianday
  public :: newJulianday
  public :: deallocateJulianday
  public :: juliandayToString
  !
  !> provides a string length for toString 
  integer, parameter :: max_julianday_str_len = 32   
  !
  type, bind(c) :: julianday
    integer(c_int64_t) :: day  !< the actual Julian day
    integer(c_int64_t) :: ms   !< the milisecond on that particular day
  end type julianday
  !
  !> @cond DOXYGEN_IGNORE_THIS
  interface
     function my_newjulianday(day, ms) result(c_pointer) bind(c, name='newJulianDay')
      import :: c_int64_t, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int64_t), value :: day
      integer(c_int64_t), value :: ms
    end function my_newjulianday
    !
    subroutine my_deallocatejulianday(jd) bind(c,name='deallocateJulianDay')
      import :: c_ptr
      type(c_ptr), value :: jd
    end subroutine my_deallocatejulianday
    !
    function my_juliandaytostring(my_julianday, string) result(string_ptr) bind(c, name='juliandayToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_julianday
      character(c_char), dimension(*) :: string
    end function my_juliandaytostring
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !
  !>
  !! @brief construct a new Julian date 
  !!
  !! @param[in] day            the Julian day
  !! @param[in] ms             an integer denoting the actual milli seconds of a day 
  !! @param[out]       errno       optional, error message 
  !! @return    ret_julianday  a pointer of type(julianday)
  function newJulianday(day, ms, errno) result(ret_julianday) !OK-TESTED.
    type(julianday), pointer :: ret_julianday
    integer(c_int64_t), intent(in) :: day
    integer(c_int64_t), intent(in) :: ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newjulianday(day, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 1 * 100 + 1
    call c_f_pointer(c_pointer, ret_julianday)
  end function newJulianday
  !
  !>
  !! @brief destructor for a Julian date 
  !!
  !! @param     my_julianday   a pointer of type(julianday)
  subroutine deallocateJulianday(my_julianday) !OK-TESTED.
    type(julianday), pointer :: my_julianday
    call my_deallocatejulianday(c_loc(my_julianday))
    my_julianday => null()
  end subroutine deallocateJulianday
  !
  !>
  !! @brief get Julian day as a string.
  !!
  !! @param[in]  my_julianday   a pointer to type(julianday). The Julian day to be converted to a string 
  !! @param[out] string         the Julian day verbose
  !! @param[out]       errno       optional, error message 
  subroutine juliandayToString(my_julianday, string, errno) !OK-TESTED.
    type(julianday), pointer :: my_julianday
    character(len=max_julianday_str_len), intent(out) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_juliandaytostring(c_loc(my_julianday), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 1 * 100 + 3
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine juliandayToString
  !
end module mtime_julianday
!>
!! @brief Date and some operations supported on Date.
!!
!! @details
!!
!___________________________________________________________________________________________________________
module mtime_date
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int64_t, c_char, c_ptr, c_null_char, &
                                       & c_loc, c_f_pointer, c_associated
  !
  implicit none
  !
  private
  !
  public :: max_date_str_len
  public :: date
  public :: newdate
  public :: deallocateDate
  public :: replaceDate
  public :: dateToString
  public :: dateToPosixString
  !
  !> provides a string length for toString 
  integer, parameter :: max_date_str_len = 32
  !
  type, bind(c) :: date
    integer(c_int64_t) :: year
    integer(c_int) :: month
    integer(c_int) :: day
  end type date
  !
  interface newdate
    module procedure newdatefromstring
    module procedure newdatefromraw
    module procedure newdatefromraw_yi8
    module procedure newdatefromconstructandcopy
  end interface newdate
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  interface
    !
    function my_newdatefromstring(string) result(c_pointer) bind(c, name='newDate')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newdatefromstring
    !
    function my_newrawdate(year, month, day) result(c_pointer) bind(c, name='newRawDate')
      import :: c_int64_t, c_int, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int64_t), value :: year
      integer(c_int), value :: month, day
    end function my_newrawdate
    !
    function my_constructandcopydate(d) result(c_pointer) bind(c,name='constructAndCopyDate')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: d
    end function my_constructandcopydate
    !
    subroutine my_deallocatedate(d) bind(c, name='deallocateDate')
      import :: c_ptr
      type(c_ptr), value :: d
    end subroutine my_deallocatedate
    !
    function my_replacedate(src, dest) result(ret_dest) bind(c, name='replaceDate')
      import :: c_ptr
      type(c_ptr) :: ret_dest
      type(c_ptr), value :: src , dest
    end function my_replacedate
    !
    function my_datetostring(my_date, string) result(string_ptr) bind(c, name='dateToString') 
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_date
      character(c_char), dimension(*) :: string
    end function my_datetostring
    !
    function my_datetoposixstring(my_date, string, fmtstr) result(string_ptr) bind(c, name='dateToPosixString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_date
      character(c_char), dimension(*) :: string
      character(c_char), dimension(*) :: fmtstr
    end function my_datetoposixstring
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS  
  !
contains
  !
  !>
  !! @brief construct a new date 
  !!
  !! @param[in] string         an ISO 8601 conforming date string
  !! @param[out]       errno       optional, error message 
  !! @return    ret_date       a pointer of type(date)
  function newdatefromstring(string,errno) result(ret_date)  !OK-TESTED.
    type(date), pointer :: ret_date
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional :: errno
    if (present(errno)) errno = 0
    c_pointer = my_newdatefromstring(trim(adjustl(string))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 1
    call c_f_pointer(c_pointer, ret_date)
  end function newdatefromstring
  !>
  !! @brief construct a new date from raw date
  !!
  !! @param[in] year           the year
  !! @param[in] month          the month
  !! @param[in] day            the day
  !! @param[out]       errno       optional, error message   
  !! @return    ret_date       a pointer of type(date)
  function newdatefromraw_yi8(year, month, day, errno) result(ret_date) !OK-TESTED.
    type(date), pointer :: ret_date
    integer(c_int64_t), intent(in) :: year
    integer(c_int), intent(in) :: month, day
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdate(year, month, day)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 2
    call c_f_pointer(c_pointer, ret_date)
  end function newdatefromraw_yi8
  !>
  !! @brief construct a new date from raw date
  !!
  !! @param[in] year           the year
  !! @param[in] month          the month
  !! @param[in] day            the day
  !! @param[out]       errno       optional, error message   
  !! @return    ret_date       a pointer of type(date)
  function newdatefromraw(year, month, day, errno) result(ret_date) !OK-TESTED.
    type(date), pointer :: ret_date
    integer(c_int), intent(in) :: year, month, day
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdate(int(year,c_int64_t), month, day)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 3
    call c_f_pointer(c_pointer, ret_date)
  end function newdatefromraw
  !>
  !! @brief construct a new date from an existing by construct and copy
  !!
  !! @param[in] src            a pointer of type(date)
  !! @param[out]       errno       optional, error message 
  !! @return    ret_date       a pointer of type(date)
  function newdatefromconstructandcopy(src, errno) result(dest) !OK-TESTED
    type(date), pointer :: dest
    type(date), target :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopydate(c_loc(src))
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 4
    call c_f_pointer(c_pointer, dest)
  end function newdatefromconstructandcopy
  !>
  !! @brief destructor for a date 
  !!
  !! @param[in] my_date        a pointer of type(date)
  subroutine deallocateDate(my_date) !OK-TESTED.
    type(date), pointer :: my_date
    call my_deallocatedate(c_loc(my_date))
    my_date => null()
  end subroutine deallocateDate
  !>
  !! @brief repace an existing date by a given one
  !!
  !! @param[in]  src            a pointer of type(date)
  !! @param[out] dest           a pointer of type(date)
  !! @param[out] errno          optional, error message 
  subroutine replaceDate(dest, src, errno)  !OK-TESTED.
    type(date), target, intent(inout) :: dest
    type(date), target, intent(in) :: src
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_replacedate(c_loc(src), c_loc(dest))
    if ((.not. c_associated(dummy_ptr)) .and. present(errno)) errno = 2 * 100 + 6
  end subroutine replaceDate
  !>
  !! @brief Get Date as an extended string.
  !!
  !! DateToString returns a string in IS08601 compliant (and extended) format.
  !!
  !! @param[in]   my_date
  !!         A pointer to type date. The date to be converted to string.
  !!
  !! @param[out]  string
  !!         String where date is to be written.
  !! 
  !! @param[out]  errno
  !!         Optional, error message   
  subroutine dateToString(my_date, string, errno) !OK-TESTED.
    type(date), pointer :: my_date
    character(len=max_date_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetostring(c_loc(my_date), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 2 * 100 + 7
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine dateToString
  !>
  !! @brief Get date and return as a string.
  !!
  !! Only dates between and including 1582-10-15 TO 9999-12-31 supported.
  !!
  !! @param[in]  my_date
  !!         A pointer to type date. The date to be converted to string.
  !!
  !! @param[out]  string
  !!         String where date is to be written.
  !!
  !! @param[in]  fmtstr
  !!         Desired Format string. CRITICAL: Inappropriate fmt string will cause dump.
  !! 
  !! @param[out]       errno       optional, error message   
  subroutine dateToPosixString(my_date, string, fmtstr, errno) !OK-TESTED.
    type(date), pointer :: my_date
    character(len=max_date_str_len) :: string
    character(len=*) :: fmtstr
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetoposixstring(c_loc(my_date), string, fmtstr)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 2 * 100 + 8
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine dateToPosixString
  !
end module mtime_date
!>
!! @brief Time and some operations supported on Time
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_time
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_char, c_ptr, c_null_char, &
                                       & c_loc, c_f_pointer, c_associated
  !
  implicit none
  !
  private
  !
  public :: max_time_str_len
  public :: time
  public :: newtime
  public :: deallocateTime
  public :: replaceTime
  public :: timeToString
  public :: timeToPosixString
  !
  !> provides a string length for toString 
  integer, parameter :: max_time_str_len = 32
  !
  type, bind(c) :: time
    integer(c_int) :: hour
    integer(c_int) :: minute
    integer(c_int) :: second
    integer(c_int) :: ms
  end type time
  !
  interface newtime
    module procedure newtimefromstring
    module procedure newtimefromraw
    module procedure newtimefromconstructandcopy
  end interface newtime
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  interface
    !
    function my_newtimefromstring(string) result(c_pointer) bind(c, name='newTime')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newtimefromstring
    !
    function my_newrawtime(hour, minute, second, ms) result(c_pointer) bind(c, name='newRawTime')
      import :: c_int, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int), value :: hour, minute, second, ms
    end function my_newrawtime
    !
    function my_constructandcopytime(t) result(c_pointer) bind(c,name='constructAndCopyTime')
    import :: c_ptr
    type(c_ptr) :: c_pointer
    type(c_ptr), value :: t
  end function my_constructandcopytime
    !
    subroutine my_deallocatetime(t) bind(c,name='deallocateTime') 
      import :: c_ptr
      type(c_ptr), value :: t
    end subroutine my_deallocatetime
    !
    function my_replacetime(src, dest) result(ret_dest) bind(c, name='replaceTime')
      import :: c_ptr
      type(c_ptr) :: ret_dest
      type(c_ptr), value :: src , dest
    end function my_replacetime
    !
    function my_timetostring(my_time, string) result(string_ptr) bind(c, name='timeToString') 
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
    end function my_timetostring
    !
    function my_timetoposixstring(my_time, string, fmtstr) result(string_ptr) bind(c, name='timeToPosixString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
      character(c_char), dimension(*) :: fmtstr
    end function my_timetoposixstring
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !
  !! @param[out]       errno       optional, error message   
  function newtimefromstring(string, errno) result(ret_time) !OK-TESTED.
    type(time), pointer :: ret_time
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newtimefromstring(trim(adjustl(string))//c_null_char)
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 3 * 100 + 1
    call c_f_pointer(c_pointer, ret_time)
  end function newtimefromstring
  !
  !! @param[out]       errno       optional, error message   
  function newtimefromraw(hour, minute, second, ms, errno) result(ret_time) !OK-TESTED.
    type(time), pointer :: ret_time
    integer(c_int), intent(in) :: hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawtime(hour, minute, second, ms)
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 3 * 100 + 2
    call c_f_pointer(c_pointer, ret_time)
  end function newtimefromraw
  !
  !! @param[out]       errno       optional, error message   
  function newtimefromconstructandcopy(src, errno) result(dest) !OK-TESTED.
    type(time), pointer :: dest
    type(time), target :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopytime(c_loc(src))
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 3 * 100 + 3
    call c_f_pointer(c_pointer, dest)
  end function newtimefromconstructandcopy
  !>
  !! @brief Destructor of Time.
  !!
  !! @param  my_time
  !!         A pointer to type time. my_time is deallocated.
  subroutine deallocateTime(my_time) !OK-TESTED.
    type(time), pointer :: my_time
    call my_deallocatetime(c_loc(my_time))
    my_time => null()
  end subroutine deallocateTime
  !>
  !! @brief COPY a time object.
  !!
  !! Routine replaceTime copies the contents of source Time into a Destination Time object.
  !!
  !! @param[in]  src
  !!         A pointer to type time. Copy "FROM" time object.
  !!
  !! @param[out]  dest
  !!      A pointer to type time. Copy "TO" time object.
  !!
  !! @param[out]       errno       optional, error message   
  subroutine replaceTime(dest,src,errno) !OK-TESTED.
    type(time), target, intent(in) :: src
    type(time), target, intent(inout) :: dest
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_replacetime(c_loc(src), c_loc(dest))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 3 * 100 + 5
  end subroutine replaceTime
  !>
  !! @brief Get time as an extended string.
  !!
  !! timetoString returns a string in IS08601 compliant (and extended) format.
  !!
  !! @param[in]  my_time
  !!         A pointer to type time. The time to be converted to string.
  !!
  !! @param[out]  string
  !!         String where time is to be written.
  !!
  !! @param[out]       errno       optional, error message   
  subroutine timeToString(my_time, string, errno) !OK-TESTED.
    type(time), pointer :: my_time
    character(len=max_time_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_timetostring(c_loc(my_time), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 3 * 100 + 6
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine timeToString
  !>
  !! @brief Get time as a Posix formated string.
  !!
  !! @param[in]  my_time
  !!         A pointer to type time. The time to be converted to string.
  !!
  !! @param[out]  string
  !!         String where time is to be written.
  !!
  !! @param[in]  fmtstr
  !!         Desired Format string. CRITICAL: Inappropriate fmt string will cause dump.
  !!
  !! @param[out]       errno       optional, error message   
  subroutine timeToPosixString(my_time, string, fmtstr, errno) !OK-TESTED.
    type(time), pointer :: my_time
    character(len=max_time_str_len) :: string
    character(len=32) :: fmtstr
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_timetoposixstring(c_loc(my_time), string, fmtstr)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 3 * 100 + 7
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine timeToPosixString
  !
end module mtime_time
!>
!! @brief DateTime and some operations supported on DateTime.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_datetime
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int64_t, c_char, c_null_char, c_ptr, &
                                       & c_loc, c_f_pointer, c_associated
  !
  use mtime_julianday
  use mtime_date
  use mtime_time
  !
  implicit none
  !
  private
  !
  public :: max_datetime_str_len
  public :: datetime
  public :: newDatetime
  public :: deallocateDatetime
  public :: datetimeToString
  public :: datetimeToPosixString
  public :: getNoOfDaysInMonthDateTime
  public :: getNoOfDaysInYearDateTime
  public :: getDayOfYearFromDateTime
  public :: getNoOfSecondsElapsedInMonthDateTime
  public :: getNoOfSecondsElapsedInDayDateTime
  public :: getJulianDayFromDatetime
  public :: getDatetimeFromJulianDay
  !
  public :: min, max
  public :: assignment(=)
  public :: operator(>)
  public :: operator(<)
  public :: operator(<=)
  public :: operator(>=)
  public :: operator(==)
  public :: operator(/=)
  !
  !> provides a string length for toString 
  integer, parameter :: max_datetime_str_len = 32
  !
  type, bind(c) :: datetime
    type(date) :: date
    type(time) :: time
  end type datetime

  !
  interface newDatetime
    module procedure newdatetimefromstring
    module procedure newdatetimefromraw
    module procedure newdatetimefromraw_yi8
    module procedure newdatetimefromconstructandcopy
  end interface newDatetime

  !
  interface min
     module procedure datetime_min
  end interface min
  !
  interface max
     module procedure datetime_max
  end interface max
  !
  interface assignment (=)
    module procedure replacedatetime
  end interface assignment (=)
  !
  interface operator (>)
    module procedure datetime_gt
  end interface operator (>)
  !
  interface operator (<)
    module procedure datetime_lt
  end interface operator (<)
  !
  interface operator (<=)
    module procedure datetime_lt_or_eq
  end interface operator (<=)
  !
  interface operator (>=)
    module procedure datetime_gt_or_eq
  end interface operator (>=)
  !
  interface operator (==)
    module procedure datetime_eq
  end interface operator (==)
  !
  interface operator (/=)
    module procedure datetime_ne
  end interface operator (/=)
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  interface
    !
    function my_newdatetime(string) result(c_pointer) bind(c, name='newDateTime')
      import :: c_ptr, c_char
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newdatetime
    !
    function my_newrawdatetime(year, month, day, hour, minute, second, ms) result(c_pointer) &
            &             bind(c, name='newRawDateTime')
      import :: c_int64_t, c_int, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int64_t), value :: year
      integer(c_int), value :: month, day, hour, minute, second, ms
    end function my_newrawdatetime
    !
    function my_constructandcopydatetime(dt) result(c_pointer) bind(c,name='constructAndCopyDateTime')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: dt
    end function my_constructandcopydatetime
    !
    subroutine my_deallocatedatetime(dt) bind(c,name='deallocateDateTime')
      import :: c_ptr
      type(c_ptr), value :: dt
    end subroutine my_deallocatedatetime
    !
    function my_comparedatetime(op1, op2) result(ret) bind(c, name='compareDatetime')
      import :: c_ptr, c_int
      integer(c_int) :: ret
      type(c_ptr), value :: op1
      type(c_ptr), value :: op2
    end function my_comparedatetime

    function my_replacedatetime(src, dest) result(ret_dest) bind(c, name='replaceDatetime')
      import :: c_ptr
      type(c_ptr) :: ret_dest
      type(c_ptr), value :: src
      type(c_ptr), value :: dest
    end function my_replacedatetime
    !
    function my_datetimetostring(my_time, string) result(string_ptr) bind(c, name='datetimeToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
    end function my_datetimetostring
    !
    function my_datetimetoposixstring(my_time, string, fmtstr) result(string_ptr) bind(c, name='datetimeToPosixString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
      character(c_char), dimension(*) :: fmtstr
    end function my_datetimetoposixstring
    !
    function my_getnoofdaysinmonthdatetime(dt) bind(c, name='getNoOfDaysInMonthDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getnoofdaysinmonthdatetime
      type(c_ptr), value :: dt
    end function my_getnoofdaysinmonthdatetime
    !
    function my_getnoofdaysinyeardatetime(dt) bind(c, name='getNoOfDaysInYearDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getnoofdaysinyeardatetime
      type(c_ptr), value :: dt
    end function my_getnoofdaysinyeardatetime
    !
    function my_getdayofyearfromdatetime(dt) bind(c, name='getDayOfYearFromDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getdayofyearfromdatetime
      type(c_ptr), value :: dt
    end function my_getdayofyearfromdatetime
    !
    function my_getnoofsecondselapsedinmonthdatetime(dt) bind(c, name='getNoOfSecondsElapsedInMonthDateTime')
      import :: c_int64_t, c_ptr
      integer(c_int64_t) :: my_getnoofsecondselapsedinmonthdatetime
      type(c_ptr), value :: dt
    end function my_getnoofsecondselapsedinmonthdatetime
    !
    function my_getnoofsecondselapsedindaydatetime(dt) bind(c, name='getNoOfSecondsElapsedInDayDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getnoofsecondselapsedindaydatetime
      type(c_ptr), value :: dt
    end function my_getnoofsecondselapsedindaydatetime
    !
    function my_getjuliandayfromdatetime(dt, jd) bind(c, name='getJulianDayFromDateTime')
      import :: c_ptr
      type(c_ptr) :: my_getjuliandayfromdatetime
      type(c_ptr), value :: dt, jd
    end function my_getjuliandayfromdatetime
    !
    function my_getdatetimefromjulianday(jd, dt) bind(c, name='getDateTimeFromJulianDay')
      import :: c_ptr
      type(c_ptr) :: my_getdatetimefromjulianday
      type(c_ptr), value :: jd, dt 
    end function my_getdatetimefromjulianday
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !
  !! @param[out]       errno       optional, error message   
  function newdatetimefromstring(string, errno) result(ret_datetime) !OK-TESTED
    type(datetime), pointer :: ret_datetime
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newdatetime(trim(adjustl(string))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 1
    call c_f_pointer(c_pointer, ret_datetime)
  end function newdatetimefromstring
  !
  !! @param[out]       errno       optional, error message   
  function newdatetimefromraw_yi8(year, month, day, hour, minute, second, ms, errno) result(ret_datetime) !OK-TESTED
    type(datetime), pointer :: ret_datetime
    integer(c_int64_t), intent(in) :: year
    integer(c_int), intent(in) :: month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdatetime(year, month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 2
    call c_f_pointer(c_pointer, ret_datetime)
  end function newdatetimefromraw_yi8
  !
  !! @param[out]       errno       optional, error message   
  function newdatetimefromraw(year, month, day, hour, minute, second, ms, errno) result(ret_datetime) !OK-TESTED
    type(datetime), pointer :: ret_datetime
    integer(c_int), intent(in) :: year, month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdatetime(int(year,c_int64_t), month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 3
    call c_f_pointer(c_pointer, ret_datetime)
  end function newdatetimefromraw
  !
  !! @param[out]       errno       optional, error message   
  function newdatetimefromconstructandcopy(src, errno) result(dest) !OK-TESTED.
    type(datetime), pointer :: dest
    type(datetime), target :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopydatetime(c_loc(src))
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 4
    call c_f_pointer(c_pointer, dest)
  end function newdatetimefromconstructandcopy


  ! @return Minimum of two date time. In case of equality we return @p a.
  !
  ! @note This function does not return a copy of one of the arguments
  ! but returns only the corresponding result point to avoid
  ! unnecessary deallocate calls.
  function datetime_min(a,b) result(res)
    type(datetime), pointer :: res
    type(datetime), pointer :: a,b

    if (a > b) then
       res => b
    else
       res => a
    end if
  end function datetime_min

  ! @return Maximum of two date time. In case of equality we return @p a.
  !
  ! @note This function does not return a copy of one of the arguments
  ! but returns only the corresponding result point to avoid
  ! unnecessary deallocate calls.
  function datetime_max(a,b) result(res)
    type(datetime), pointer :: res
    type(datetime), pointer :: a,b

    if (a < b) then
       res => b
    else
       res => a
    end if
  end function datetime_max

  !>
  !! @brief Destructor of DateTime.
  !!
  !! @param  my_datetime
  !!         A pointer to type datetime. my_datetime is deallocated.
  subroutine deallocateDatetime(my_datetime) !OK-TESTED.
    type(datetime), pointer :: my_datetime
    call my_deallocatedatetime(c_loc(my_datetime))
    nullify(my_datetime)
  end subroutine deallocateDatetime
  !>
  !! @brief Get DateTime as a string.
  !!
  !! datetimeToString returns a string in IS08601 compliant (and extended) format.
  !!
  !! @param[in]  my_datetime
  !!         A pointer to struct _datetime. The datetime to be converted to string.
  !!
  !! @param[out]  string
  !!         String where datetime is to be written.
  !!
  !! @param[out]       errno       optional, error message   
  subroutine datetimeToString(my_datetime, string, errno) !OK-TESTED
    TYPE(datetime), target, intent(in) :: my_datetime
    character(len=max_datetime_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetimetostring(c_loc(my_datetime), string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 4 * 100 + 6
      string ='<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine datetimeToString
  !>
  !! @brief Get DateTime in 'struct tm' format and return as a string.
  !!
  !! Only dates between and including 1582-10-15 TO 9999-12-31 supported.
  !!
  !! @param  my_datetime
  !!         An object of type datetime. The datetime to be converted to string.
  !!
  !! @param  string
  !!         String where datetime is to be written.
  !!
  !! @param  fmtstr
  !!         Desired Format string. CRITICAL: Inappropriate fmt string will cause dump.
  !! 
  !! @param[out]       errno       optional, error message   
  subroutine datetimeToPosixString(my_datetime, string, fmtstr, errno) !OK-TESTED.
    type(datetime), target, intent(in) :: my_datetime
    character(len=max_datetime_str_len) :: string
    character(len=*) :: fmtstr
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetimetoposixstring(c_loc(my_datetime), string, fmtstr)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 4 * 100 + 7
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine datetimeToPosixString
  !
  ! NOTE: Do not call the function using replacedatetime(.,.) directly; Use overloaded '=' instead as in "dest = src"
  subroutine replacedatetime(dest, src) !OK-TESTED. 
    type(datetime), target, intent(inout) :: dest
    type(datetime), target, intent(in) :: src
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_replacedatetime(c_loc(src), c_loc(dest))
  end subroutine replacedatetime
  !
  function datetime_gt(op1, op2) result(gt) !OK-TESTED.
    logical :: gt
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret == 1) then
      gt = .true.
    else
      gt = .false.
    endif
  end function datetime_gt
  !
  function datetime_lt(op1, op2) result(lt) !OK-TESTED.
    logical :: lt
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret == -1) then
      lt = .true.
    else
      lt = .false.
    endif
  end function datetime_lt
  !
  function datetime_lt_or_eq(op1, op2) result(lt_or_eq) !OK-TESTED.
    logical :: lt_or_eq
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == -1)) then
      lt_or_eq = .true.
    else
      lt_or_eq = .false.
    endif
  end function datetime_lt_or_eq
  !
  function datetime_gt_or_eq(op1, op2) result(gt_or_eq) !OK-TESTED
    logical :: gt_or_eq
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == 1)) then
      gt_or_eq = .true.
    else
      gt_or_eq = .false.
    endif
  end function datetime_gt_or_eq
  !
  function datetime_eq(op1, op2) result(eq) !OK-TESTED.
    logical :: eq
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret == 0) then
      eq = .true.
    else
      eq = .false.
    endif
  end function datetime_eq
  !
  function datetime_ne(op1, op2) result(ne) !OK-TESTED.
    logical :: ne
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret /= 0) then
      ne = .true.
    else
      ne = .false.
    endif
  end function datetime_ne
  !>
  !! @brief Get nod (number of Days) in the month of DateTime.
  !!
  !! Routine getNoOfDaysInMonthDateTime returns number of days in the month of DateTime. This routine  
  !! supports all calendar types.
  !!
  !! For eg. the number of days for 2001-10-15T00:00:00.000 will be 31 for Gregorian Calendar. 
  !! Similarly, this value will be 30 for Calendar of type 360 day-Calendar.
  !!
  !! @param[in]  dt
  !!         A pointer to type datetime.
  !!
  !! @return nod
  !!         Integer value of nod. The value depends on the month and the calendar type. Zero indicates error.
  !!
  !! @param[out]       errno       optional, error message   
  function getNoOfDaysInMonthDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getNoOfDaysInMonthDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfDaysInMonthDateTime = my_getnoofdaysinmonthdatetime(c_loc(dt))
    if (getNoOfDaysInMonthDateTime == 0 .and. present(errno)) errno = 4 * 100 + 15
  end function getNoOfDaysInMonthDateTime
  !>
  !! @brief Get number of days in the Year of DateTime.
  !!
  !! Routine getNoOfDaysInYearDateTime returns number of days in the Year of DateTime. This routine  
  !! supports all calendar types.
  !!
  !! Number of days returned will depend on the calendar type and if applicable, leap v/s non leap year.
  !!
  !! @param[in]  dt
  !!         A pointer to type datetime.
  !!
  !! @return nod
  !!         Integer value of nod. The value depends on the year and the calendar type. Zero indicates error.
  !!
  !! @param[out]       errno       optional, error message   
  function getNoOfDaysInYearDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getNoOfDaysInYearDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfDaysInYearDateTime = my_getnoofdaysinyeardatetime(c_loc(dt))
    if (getNoOfDaysInYearDateTime == 0 .and. present(errno)) errno = 4 * 100 + 16
  end function getNoOfDaysInYearDateTime
  !>
  !! @brief Get the 'day-of-year' value of a DateTime.
  !!
  !! Routine getDayOfYearFromDateTime returns Day of Year for the DateTime. This routine supports
  !! all Calendar types.
  !!
  !! For eg. the day of year value for 2001-10-15T00:00:00.000 will be 288 for Gregorian Calendar. 
  !! Similarly, this value will be 285 for Calendar of type 360 day-Calendar.
  !!
  !! @param[in]  dt
  !!         A pointer to type datetime. Retrieve the 'day-of-year' from this DT object.
  !!
  !! @return doy
  !!         Integer value of doy. The value depends on the calendar type. Zero indicates error.
  !!
  !! @param[out]       errno       optional, error message   
  function getDayOfYearFromDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getDayOfYearFromDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getDayOfYearFromDateTime = my_getdayofyearfromdatetime(c_loc(dt))
    if (getDayOfYearFromDateTime == 0 .and. present(errno)) errno = 4 * 100 + 17
  end function getDayOfYearFromDateTime
  !>
  !! @brief Get number of seconds elapsed in the month of DateTime.
  !!
  !! @param[in] dt
  !!         A pointer to type datetime.
  !!
  !! @return no_of_seconds
  !!         int(i8) value of no_of_seconds. -1 indicates error.
  !!
  !! @param[out]       errno       optional, error message   
  function getNoOfSecondsElapsedInMonthDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int64_t) :: getNoOfSecondsElapsedInMonthDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfSecondsElapsedInMonthDateTime = my_getnoofsecondselapsedinmonthdatetime(c_loc(dt))
    if (getNoOfSecondsElapsedInMonthDateTime == -1 .and. present(errno)) errno = 4 * 100 + 18
  end function getNoOfSecondsElapsedInMonthDateTime
  !>
  !! @brief Get number of seconds elapsed in the day of DateTime.
  !!
  !! @param[in]  dt
  !!         A pointer to type datetime.
  !!
  !! @return no_of_seconds
  !!         int value of no_of_seconds. -1 indicates error.
  !!
  !! @param[out]       errno       optional, error message   
  function getNoOfSecondsElapsedInDayDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getNoOfSecondsElapsedInDayDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfSecondsElapsedInDayDateTime = my_getnoofsecondselapsedindaydatetime(c_loc(dt))
    if (getNoOfSecondsElapsedInDayDateTime == -1 .and. present(errno)) errno = 4 * 100 + 19
  end function getNoOfSecondsElapsedInDayDateTime
  !>
  !! @brief Get the Julian Day from DateTime.
  !!
  !! The routine getJulianDayFromDateTime returns the equivalent Julian date to DateTime. Internally
  !! it calls translation routines based on Calendar type.
  !!
  !! @param[in]  dt
  !!         A pointer to type datetime. The DT's value is converted to julian day value.
  !!
  !! @param[out]  jd
  !!         A pointer to type julianday. JD where the converted value is stored.
  !!
  !! @param[out]       errno       optional, error message   
  subroutine getJulianDayFromDatetime(dt, jd, errno) !OK-TESTED.
    type(datetime), target :: dt
    type(julianday), pointer :: jd
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_getjuliandayfromdatetime(c_loc(dt), c_loc(jd))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 4 * 100 + 20
  end subroutine getJulianDayFromDatetime
  !>
  !! @brief Get the DateTime from Julian Day.
  !!
  !! The routine getDateTimeFromJulianDay returns the equivalent DateTime to Julian date. Internally
  !! it calls translation routines based on Calendar type.
  !!
  !! @param[in]  jd
  !!         A pointer to type julianday. The JD's value is converted to julian day value.
  !!
  !! @param[out]  dt
  !!         A pointer to type datetime. The DT where the converted value is stored.
  !!
  !! @param[out]       errno       optional, error message   
  subroutine getDatetimeFromJulianDay(jd, dt, errno)
    TYPE(julianday), TARGET, INTENT(in) :: jd
    TYPE(datetime), TARGET, INTENT(out) :: dt
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_getdatetimefromjulianday(c_loc(jd), c_loc(dt))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 4 * 100 + 21
  end subroutine getDatetimeFromJulianDay
  !
end module mtime_datetime
!>
!! @brief TimeDelta and some operations supported on TimeDelta.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_timedelta
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int32_t, c_int64_t, c_float, c_double, c_char, &
                                       & c_ptr, c_null_char, c_loc, c_f_pointer, c_associated
  !
  use mtime_datetime
  use mtime_date 
  use mtime_juliandelta
  !use mtime_time
  !
  implicit none
  !
  private
  !
  public :: max_timedelta_str_len
  public :: timedelta
  public :: divisionquotienttimespan
  public :: newTimedelta
  public :: deallocateTimedelta
  public :: getTimeDeltaFromDate
  public :: getTimeDeltaFromDateTime
  public :: getTotalMilliSecondsTimeDelta
  public :: getTotalSecondsTimeDelta
  public :: timedeltaToString
  public :: moduloTimedelta
  public :: getPTStringFromMS
  public :: getPTStringFromSeconds
  public :: getPTStringFromMinutes
  public :: getPTStringFromHours
  public :: timeDeltaToJulianDelta
  public :: divideTimeDeltaInSeconds
  public :: divideTwoDatetimeDiffsInSeconds  
  public :: divideDatetimeDifferenceInSeconds  
  public :: operator(+)
  public :: operator(-)
  public :: operator(*)
  public :: operator(>)
  public :: operator(<)
  public :: operator(<=)
  public :: operator(>=)
  public :: operator(==)
  public :: operator(/=)
  !
  !> provides a string length for toString 
  integer, parameter :: max_timedelta_str_len = 32
  !
  type, bind(c) :: timedelta
    integer(c_int) :: flag_std_form
    character(c_char) :: sign
    integer(c_int64_t) :: year
    integer(c_int) :: month
    integer(c_int) :: day
    integer(c_int) :: hour
    integer(c_int) :: minute
    integer(c_int) :: second
    integer(c_int) :: ms
  end type timedelta
  !
  type, bind(c) :: divisionquotienttimespan
    integer(c_int64_t) :: quotient;
    integer(c_int64_t) :: remainder_in_ms;
  end type divisionquotienttimespan
  !
  interface newTimedelta
    module procedure newtimedeltafromstring
    module procedure newtimedeltafromraw
    module procedure newtimedeltafromraw_yi8
    module procedure newtimedeltafromconstructandcopy
  end interface newTimedelta
  !
  interface operator (+)
    module procedure addtimedeltatodatetime
    module procedure adddatetimetotimedelta
    module procedure addtimedeltatodate
    module procedure adddatetotimedelta
    module procedure elementwiseAddTimeDeltatoTimeDelta
  end interface operator (+)
  !
  interface operator (-)
    module procedure getTimeDeltaFromDate
    module procedure getTimeDeltaFromDateTime
  end interface operator (-)
  !
  interface operator (*)
    module procedure elementwiseScalarMultiplyTimeDelta
    module procedure elementwiseScalarMultiplyTimeDeltaInv
    module procedure elementwiseScalarMultiplyTimeDelta_long
    module procedure elementwiseScalarMultiplyTimeDeltaInv_long
    module procedure elementwiseScalarMultiplyTimeDelta_real
    module procedure elementwiseScalarMultiplyTimeDeltaInv_real
  end interface operator (*)
  !
  interface operator (>)
  module procedure timedelta_gt
  end interface operator (>)
  !
  interface operator (<)
  module procedure timedelta_lt
  end interface operator (<)
  !
  interface operator (<=)
  module procedure timedelta_lt_or_eq
  end interface operator (<=)
  !
  interface operator (>=)
  module procedure timedelta_gt_or_eq
  end interface operator (>=)
  !
  interface operator (==)
  module procedure timedelta_eq
  end interface operator (==)
  !
  interface operator (/=)
  module procedure timedelta_ne
  end interface operator (/=)
  !
  interface getPTStringFromSeconds
    module procedure getPTStringFromSecondsInt
    module procedure getPTStringFromSecondsFloat
    module procedure getPTStringFromSecondsDouble
  end interface getPTStringFromSeconds
  !
  !> @cond DOXYGEN_IGNORE_THIS
  interface
    !
    function my_newtimedeltafromstring(string) result(c_pointer) bind(c, name='newTimeDelta')
      import :: c_char, c_ptr
      type(c_ptr)                     :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newtimedeltafromstring
    !
    function my_newrawtimedelta(sign, year, month, day, hour, minute, second, ms) result(c_pointer) &
            &             bind(c, name='newRawTimeDelta')
      import :: c_int64_t, c_char, c_int, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), value :: sign
      integer(c_int64_t), value :: year
      integer(c_int), value :: month, day, hour, minute, second, ms
    end function my_newrawtimedelta
    !
    function my_constructandcopytimedelta(td) result(c_pointer) bind(c,name='constructAndCopyTimeDelta')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: td
    end function my_constructandcopytimedelta
    !
    subroutine my_deallocatetimedelta(dt) bind(c, name='deallocateTimeDelta')
      import :: c_ptr
      type(c_ptr), value :: dt
    end subroutine my_deallocatetimedelta
    !
    function my_comparetimedelta(op1, op2) result(ret) bind(c, name='compareTimeDelta')
      import :: c_ptr, c_int
      integer(c_int) :: ret
      type(c_ptr), value :: op1
      type(c_ptr), value :: op2
    end function my_comparetimedelta
    !
    function my_gettimedeltafromdate(my_date1,my_date2,timedelta_return) result(timedelta_ptr) &
             &    bind(c,name='getTimeDeltaFromDate')
      import :: c_ptr
      type(c_ptr) :: timedelta_ptr
      type(c_ptr), value :: my_date1, my_date2, timedelta_return
    end function my_gettimedeltafromdate
    !
    function my_gettimedeltafromdatetime(my_datetime1,my_datetime2,timedelta_return) result(timedelta_ptr) &
             &    bind(c,name='getTimeDeltaFromDateTime')
      import :: c_ptr
      type(c_ptr) :: timedelta_ptr
      type(c_ptr), value :: my_datetime1, my_datetime2, timedelta_return
    end function my_gettimedeltafromdatetime
    !
    function my_gettotalmillisecondstimedelta(my_timedelta, my_datetime) &
            &   bind(c, name='getTotalMilliSecondsTimeDelta')
      import :: c_ptr, c_int64_t
      integer(c_int64_t) :: my_gettotalmillisecondstimedelta
      type(c_ptr), value :: my_timedelta, my_datetime
    end function my_gettotalmillisecondstimedelta
    !
    function my_gettotalsecondstimedelta(my_timedelta, my_datetime) bind(c, name='getTotalSecondsTimeDelta')
      import :: c_ptr, c_int64_t
      integer(c_int64_t) :: my_gettotalsecondstimedelta
      type(c_ptr), value :: my_timedelta, my_datetime
    end function my_gettotalsecondstimedelta
    !
    function my_timedeltatostring(td, tostring) result(string_ptr) bind(c, name='timedeltaToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: td
      character(c_char), dimension(*) :: tostring
    end function my_timedeltatostring
    !
    function my_addtimedeltatodatetime(my_datetime, my_timedelta, ret_datetime) result(datetime_ptr) &
         &   bind(c, name='addTimeDeltaToDateTime')
      import :: c_ptr
      type(c_ptr) :: datetime_ptr
      type(c_ptr), value :: my_datetime, my_timedelta, ret_datetime
    end function my_addtimedeltatodatetime
    !
    function my_addtimedeltatodate(my_date, my_timedelta, ret_date) result(date_ptr) &
         &   bind(c, name='addTimeDeltaToDate')
      import :: c_ptr
      type(c_ptr) :: date_ptr
      type(c_ptr), value :: my_date, my_timedelta, ret_date
    end function my_addtimedeltatodate
    !
    function my_elementwisescalarmultiplytimedelta(my_timedelta, lambda, scaled_timedelta) result(timedelta_return) &
         &   bind(c, name='elementwiseScalarMultiplyTimeDelta')
      import :: c_ptr, c_int64_t
      type(c_ptr) ::timedelta_return
      type(c_ptr), value :: my_timedelta, scaled_timedelta
      integer(c_int64_t), value :: lambda
    end function my_elementwisescalarmultiplytimedelta
    !
    function my_elementwisescalarmultiplytimedeltadp(my_timedelta, lambda, scaled_timedelta) RESULT(timedelta_return) &
         &   bind(c, name='elementwiseScalarMultiplyTimeDeltaDP')
      import :: c_ptr, c_double
      type(c_ptr) ::timedelta_return
      type(c_ptr), value :: my_timedelta, scaled_timedelta
      real(c_double), value :: lambda
    end function my_elementwisescalarmultiplytimedeltadp
    !
    function my_elementwiseAddTimeDeltatoTimeDelta(my_timedelta1,my_timedelta2,added_timedelta) result(timedelta_return) &
         &   bind(c, name='elementwiseAddTimeDeltatoTimeDelta')
      import :: c_ptr
      type(c_ptr) :: timedelta_return
      type(c_ptr), value :: my_timedelta1,my_timedelta2,added_timedelta
    end function my_elementwiseAddTimeDeltatoTimeDelta
    !
    function my_modulotimedelta(a, p, quot) result(rem) bind(c, name='moduloTimedelta')
      import :: c_ptr, c_int64_t
      integer(c_int64_t) :: rem 
      type(c_ptr), value :: a
      type(c_ptr), value :: p
      type(c_ptr), value :: quot      
    end function my_modulotimedelta
    !
    function my_getptstringfromms(ms, tostring) result(string_ptr) bind(c, name='getPTStringFromMS')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t), value :: ms
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromms
    !
    function my_getptstringfromsecondsint(s, tostring) result(string_ptr) bind(c, name='getPTStringFromSeconds')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t), value :: s
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromsecondsint
    !
    function my_getptstringfromsecondsfloat(s, tostring) result(string_ptr) bind(c, name='getPTStringFromSecondsFloat')
      import :: c_ptr, c_float, c_char
      type(c_ptr) :: string_ptr
      real(c_float), value :: s
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromsecondsfloat
    !
    function my_getptstringfromsecondsdouble(s, tostring) result(string_ptr) bind(c, name='getPTStringFromSecondsDouble')
      import :: c_ptr, c_double, c_char
      type(c_ptr) :: string_ptr
      real(c_double), value :: s
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromsecondsdouble
    !
    function my_getptstringfromminutes(m, tostring) result(string_ptr) bind(c, name='getPTStringFromMinutes')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t),value :: m
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromminutes
    !
    function my_getptstringfromhours(h, tostring) result(string_ptr) bind(c, name='getPTStringFromHours')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t),value :: h
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromhours
    !
    FUNCTION my_timedeltatojuliandelta(td,dt,jd) RESULT(c_pointer) bind(c,name='timeDeltaToJulianDelta') 
      import :: c_int64_t, c_ptr
      type(c_ptr) :: c_pointer
      TYPE(c_ptr), value :: td
      TYPE(c_ptr), value :: dt
      type(c_ptr), value :: jd
    END FUNCTION my_timedeltatojuliandelta
    !
    function my_divideTimeDeltaInSeconds(dividend, divisor, quotient) result(ret_quotient) bind(c,name='divideTimeDeltaInSeconds') 
      import :: c_ptr
      type(c_ptr) :: ret_quotient
      type(c_ptr), value :: dividend 
      type(c_ptr), value :: divisor
      type(c_ptr), value :: quotient
    end function my_divideTimeDeltaInSeconds
    !
    function my_divideDatetimeDifferenceInSeconds(dt1, dt2, divisor, quotient) result(ret_quotient) &
         &                                                                     bind(c,name='divideDatetimeDifferenceInSeconds') 
      import :: c_ptr
      type(c_ptr) :: ret_quotient
      type(c_ptr), value :: dt1
      type(c_ptr), value :: dt2
      type(c_ptr), value :: divisor
      type(c_ptr), value :: quotient
    end function my_divideDatetimeDifferenceInSeconds
    !
    function my_divideTwoDatetimeDiffsInSeconds(dt1_dividend, dt2_dividend, &
         &                                      dt1_divisor,  dt2_divisor, denominator, quotient) &
         &                                       result(ret_quotient) &
         &                                       bind(c,name='divideTwoDatetimeDiffsInSeconds') 
      import :: c_ptr
      type(c_ptr) :: ret_quotient
      type(c_ptr), value :: dt1_dividend
      type(c_ptr), value :: dt2_dividend
      type(c_ptr), value :: dt1_divisor
      type(c_ptr), value :: dt2_divisor
      type(c_ptr), value :: denominator
      type(c_ptr), value :: quotient
    end function my_divideTwoDatetimeDiffsInSeconds
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS  
  !
contains
  !
  !! @param[out]       errno       optional, error message   
  function newtimedeltafromstring(string, errno) result(ret_timedelta) !OK-TESTED.
    type(timedelta), pointer :: ret_timedelta
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    ret_timedelta => null()
    c_pointer = my_newtimedeltafromstring(trim(adjustl(string))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) then
      errno = 5 * 100 + 1
    else
      call c_f_pointer(c_pointer, ret_timedelta)
    endif
  end function newtimedeltafromstring
  !
  !! @param[out]       errno       optional, error message   
  function newtimedeltafromraw(sign, year, month, day, hour, minute, second, ms, errno) result(ret_timedelta) 
    type(timedelta), pointer :: ret_timedelta
    character(len=*), intent(in) :: sign
    integer(c_int), intent(in) :: year, month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    character(c_char) ::c_sign
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_sign = sign(1:1)
    c_pointer = my_newrawtimedelta(c_sign, int(year,c_int64_t), month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 5 * 100 + 2
    call c_f_pointer(c_pointer, ret_timedelta)
  end function newtimedeltafromraw
  !
  !! @param[out]       errno       optional, error message   
  function newtimedeltafromraw_yi8(sign, year, month, day, hour, minute, second, ms, errno) result(ret_timedelta)
    type(timedelta), pointer :: ret_timedelta
    character(len=*), intent(in) :: sign
    integer(c_int64_t), intent(in) :: year
    integer(c_int), intent(in) :: month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    character(c_char) ::c_sign    
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_sign = sign(1:1)
    c_pointer = my_newrawtimedelta(c_sign, year, month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 5 * 100 + 3
    call c_f_pointer(c_pointer, ret_timedelta)
  end function newtimedeltafromraw_yi8
  !
  !! @param[out]       errno       optional, error message   
  function newtimedeltafromconstructandcopy(src, errno) result(dest) !OK-TESTED.
    type(timedelta), pointer :: dest
    type(timedelta), target :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopytimedelta(c_loc(src))
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 5 * 100 + 4
    call c_f_pointer(c_pointer, dest)
  end function newtimedeltafromconstructandcopy
  !>
  !! @brief Destructor of TimeDelta.
  !!
  !! @param  my_timedelta
  !!         A pointer to typetimedelta. my_timedelta is deallocated.
  subroutine deallocateTimedelta(my_timedelta) !OK-TESTED.
    type(timedelta), pointer :: my_timedelta
    call my_deallocatetimedelta(c_loc(my_timedelta))
    nullify(my_timedelta)
  end subroutine deallocateTimedelta
  !
  function timedelta_gt(op1, op2) result(gt)
    logical :: gt
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret == 1) then
      gt = .true.
    else
      gt = .false.
    endif
  end function timedelta_gt
  !
  function timedelta_lt(op1, op2) result(lt)
    logical :: lt
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret == -1) then
      lt = .true.
    else
      lt = .false.
    endif
  end function timedelta_lt
  !
  function timedelta_lt_or_eq(op1, op2) result(lt_or_eq)
    logical :: lt_or_eq
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == -1)) then
      lt_or_eq = .true.
    else
      lt_or_eq = .false.
    endif
  end function timedelta_lt_or_eq
  !
  function timedelta_gt_or_eq(op1, op2) result(gt_or_eq)
    logical :: gt_or_eq
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == 1)) then
      gt_or_eq = .true.
    else
      gt_or_eq = .false.
    endif
  end function timedelta_gt_or_eq
  !
  function timedelta_eq(op1, op2) result(eq)
    logical :: eq
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret == 0) then
      eq = .true.
    else
      eq = .false.
    endif
  end function timedelta_eq
  !
  function timedelta_ne(op1, op2) result(ne)
    logical :: ne
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret /= 0) then
      ne = .true.
    else
      ne = .false.
    endif
  end function timedelta_ne
  !>
  !! @brief Get the TimeDelta between two Dates op1 and op2 as (op1-op22).
  !!
  !! Routine getTimeDeltaFromDate 'substracts' two Dates and returns the TimeDelta between
  !! them. Internally, Dates are converted to DateTimes and then delta is calculated using 
  !! getTimeDeltaFromDateTime().
  !! 
  !! This routine  handles all supported Calendar types; i.e. the translation from Calendar date 
  !! to Julian date and conversion from Julian Delta to normal TimeDetla is Calendar-type dependent. 
  !! For eg. for Calendar type Gregorian, the TimeDelta between 2001-02-01 and 2001-01-01 will be 1 month. 
  !! Similarly, for Calendar of type 360-Day-Calendar, the TimeDelta will be 1 month. It must be noted 
  !! however, that the two dates differ by 31 and 30 days respectively.
  !!
  !! @param  op1
  !!         A pointer to type date. 
  !!
  !! @param  op2
  !!         A pointer to type date.
  !!
  !! @return ret
  !!         A pointer to TimeDelta containing the result of substraction.
  function getTimeDeltaFromDate(op1,op2) result(ret) !OK-TESTED.
    type(timedelta), target :: ret
    type(date), target, intent(in)  :: op1, op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_gettimedeltafromdate(c_loc(op1),c_loc(op2),c_loc(ret))
  end function getTimeDeltaFromDate
  !>
  !! @brief Get the TimeDelta between two DateTimes op1 and op2 as (op1-op2).
  !!
  !! Routine getTimeDeltaFromDateTime 'substracts' two DateTime's and returns the TimeDelta between
  !! them. Each datetime is converted to an equivalent Julian Date. Substraction is then performed
  !! on Julian axis. The "Julian delta" is finally converted back to normal calendar delta. 
  !! 
  !! This routine handles all supported Calendar types; i.e. the translation from Calendar date 
  !! to Julian date and conversion from Julian Delta to normal TimeDetla is Calendar-type dependent. 
  !! For eg. for Calendar type Gregorian, the TimeDelta between 2001-02-01T00:00:00.000 and 
  !! 2001-01-01T00:00:00.000 will be 1 month. Similarly, for Calendar of type 360-Day-Calendar, 
  !! the TimeDelta will be 1 month. It must be noted however, that the two dates differ by 31 and 
  !! 30 days respectively. 
  !!
  !! @param  op1
  !!         A pointer to type datetime. 
  !!
  !! @param  op2
  !!         A pointer to type datetime.
  !!
  !! @return ret
  !!        A pointer to TimeDelta containing the result of substraction.
  function getTimeDeltaFromDateTime(op1,op2) result(ret) !OK-TESTED.
    type(timedelta), target :: ret
    type(datetime), target, intent(in) :: op1, op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_gettimedeltafromdatetime(c_loc(op1),c_loc(op2),c_loc(ret))
  end function getTimeDeltaFromDateTime
  !> 
  !! @brief Get total number of milliseconds in timedelta.
  !!
  !! Routine getTotalMilliSecondsTimeDelta returns the total number of milliseconds in TimeDelta. 
  !! Notice that TimeDelta is not uniquely defined but depends on the definition of corresponding 
  !! DateTime. TimeDelta is first converted to corresponding delta on the Julian axis. Julian delta 
  !! is finally converted to the correct millisecond value.
  !!
  !! @param[in]  td
  !!         A pointer to type timedelta. Retrieve the number of milliseconds in this TD object.
  !!
  !! @param[in]  dt
  !!         A pointer to type datetime. Reference Datetime for the TD.
  !!
  !! @param[out] errno
  !!         Optional, error message
  !!
  !! @return totalmilliSeconds
  !!         Integer value of totalmilliSeconds. 0 indicates error.
  !!
  !! WARNING: TD 0 is error. If you know your TD is 0, ignore the error flag.
  function getTotalMilliSecondsTimeDelta(td, dt, errno)  !OK-TESTED.
    integer(c_int64_t) :: getTotalMilliSecondsTimeDelta
    type(timedelta), target, intent(in):: td
    type(datetime), target, intent(in) :: dt
    integer, optional:: errno
    if (present(errno)) errno = 0
    getTotalMilliSecondsTimeDelta = my_gettotalmillisecondstimedelta(c_loc(td), c_loc(dt))
    if ((getTotalMilliSecondsTimeDelta == 0) .and. present(errno)) errno = 5 * 100 + 8
  end function getTotalMilliSecondsTimeDelta
  !>
  !! @brief Get total number of seconds in timedelta.
  !!
  !! Routine getTotalSecondsTimeDelta returns the total number of seconds in TimeDelta. Notice that TimeDelta 
  !! is not uniquely defined but depends on the definition of corresponding DateTime. Internally, number of seconds
  !! is calculated by calling the routine getTotalMilliSecondsTimeDelta() and then converting the millisecond value 
  !! to seconds by dividing it by 1000.
  !!
  !! @param[in]  td
  !!         A pointer to struct _timedelta. Retrieve the number of seconds in this TD object.
  !!
  !! @param[in]  dt
  !!         A pointer to struct _datetime. Reference Datetime for the TD.
  !!
  !! @param[out] errno
  !!         Optional, error message
  !!
  !! @return totalSeconds
  !!         Integer value of totalSeconds. 0 indicates error.
  !!
  !! WARNING: TD 0 is error. If you know your TD is 0, ignore the error flag.
  function getTotalSecondsTimeDelta(td, dt, errno) !OK-TESTED.
    integer(c_int64_t) :: getTotalSecondsTimeDelta
    type(timedelta), target, intent(in) :: td
    type(datetime), target, intent(in) :: dt
    integer, optional:: errno
    if (present(errno)) errno = 0
    getTotalSecondsTimeDelta = my_gettotalsecondstimedelta(c_loc(td), c_loc(dt))
    ! error handling with "errno" not yet implemented.
  end function getTotalSecondsTimeDelta
  !>
  !! @brief Get TimeDelta as an extended string.
  !!
  !! timedeltaToString returns a string in IS08601 compliant (and extended) format.
  !!
  !! @param[in] my_timedelta
  !!         A pointer to type timedelta. The timedelta to be converted to string.
  !!
  !! @param[out] string
  !!         String where timedelta is to be written.
  !!
  !! @param[out]       errno       optional, error message     
  subroutine timedeltaToString(my_timedelta, string, errno) !OK-TESTED.
    type(timedelta), pointer :: my_timedelta
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    string(:) = ''
    dummy_ptr = my_timedeltatostring(c_loc(my_timedelta), string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 10
      string ='<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine timedeltaToString
  !
  function addTimedeltaToDatetime(op1, op2) result(ret) !OK-TESTED.
    type(datetime), target :: ret
    type(datetime), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodatetime(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addTimedeltaToDatetime
  !
  function addDatetimeToTimedelta(op2, op1) result(ret) !OK-TESTED.
    type(datetime), target :: ret
    type(datetime), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodatetime(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addDatetimeToTimedelta
  !
  function addTimedeltaToDate(op1, op2) result(ret) !OK-TESTED.
    type(date), target :: ret
    type(date), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodate(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addTimedeltaToDate
  !
  function addDateToTimedelta(op2, op1) result(ret) !OK-TESTED.
    type(date), target :: ret
    type(date), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodate(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addDateToTimedelta  
  !
  function elementwiseScalarMultiplyTimeDelta(base_td, ilambda) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int32_t), intent(in) :: ilambda
    type(timedelta), target, intent(in) :: base_td
    integer(c_int64_t) :: lambda
    type(c_ptr) :: dummy_ptr
    lambda = int(ilambda, c_int64_t)
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDelta
  !
  function elementwiseScalarMultiplyTimeDeltaInv(ilambda, base_td) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int32_t), intent(in) :: ilambda
    type(timedelta), target, intent(in) :: base_td
    integer(c_int64_t) :: lambda
    type(c_ptr) :: dummy_ptr
    lambda = int(ilambda, c_int64_t)
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDeltaInv
  !
  function elementwiseScalarMultiplyTimeDelta_long(base_td, lambda) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int64_t), intent(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDelta_long
  !
  function elementwiseScalarMultiplyTimeDeltaInv_long(lambda, base_td) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int64_t), intent(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDeltaInv_long
  !
  function elementwisescalarmultiplytimedelta_real(base_td, lambda) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    real(c_double), intent(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedeltadp(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwisescalarmultiplytimedelta_real
  !
  function elementwisescalarmultiplytimedeltaInv_real(lambda, base_td) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    real(c_double), intent(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedeltadp(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwisescalarmultiplytimedeltaInv_real
  !
  function elementwiseAddTimeDeltatoTimeDelta(td1, td2) result(added_td) !OK-TESTED.
    type(timedelta), target :: added_td
    type(timedelta), target, intent(in) :: td1, td2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwiseaddtimedeltatotimedelta(c_loc(td1), c_loc(td2), c_loc(added_td))
  end function elementwiseAddTimeDeltatoTimeDelta
  !>
  !! @brief Returns modulo(a,p) and the quotient.
  !!
  !! @param[in]  a
  !!         A pointer to type timedelta.
  !!
  !! @param[in]  p
  !!         A pointer to type timedelta.
  !!
  !! @param[out]  quot
  !!         The quotient of a divided by p.
  !!
  !! @return rem
  !!       modulo(a, p)
  function moduloTimedelta(a, p, quot) result(rem)
    type(timedelta), target, intent(in) :: a
    type(timedelta), target, intent(in) :: p
    integer(c_int64_t), target, intent(out) :: quot
    integer(c_int64_t) :: rem
    rem = my_modulotimedelta(c_loc(a), c_loc(p), c_loc(quot))
  end function moduloTimedelta
  !>
  !! @brief Return a PT String corresponding to arbitrary number of milliseconds.
  !!
  !! getPTStringFromMS() translates ms values to ISO 8601 compliant timedelta string. 
  !! Conversion of ms >= 86400000 and  ms <= -86400000 not supported.
  !!
  !! @param[in]  ms
  !!         An int64_t value to be translated.
  !!
  !! @param[out]  string
  !!         Translated string is written here.
  !!
  subroutine getPTStringFromMS(ms, string, errno) !OK-TESTED.
    integer(c_int64_t), intent(in) :: ms
    character(len=*) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    string(:) = ''
    dummy_ptr = my_getptstringfromms(ms, string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 11
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine getPTStringFromMS
  !
  subroutine getPTStringFromSecondsInt(s, string, errno) !OK-TESTED.
    integer(c_int64_t) :: s
    character(len=*) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    string(:) = ''
    if (present(errno)) errno = 0
    dummy_ptr = my_getptstringfromsecondsint(s, string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 11
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine getPTStringFromSecondsInt
  !
  subroutine getPTStringFromSecondsFloat(s, string, errno) !OK-TESTED.
    real(c_float) :: s
    character(len=*) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    string(:) = ''
    if (present(errno)) errno = 0
    dummy_ptr = my_getptstringfromsecondsfloat(s, string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 11
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine getPTStringFromSecondsFloat
  !
  subroutine getPTStringFromSecondsDouble(s, string, errno) !OK-TESTED.
    real(c_double) :: s
    character(len=*) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    string(:) = ''
    if (present(errno)) errno = 0
    dummy_ptr = my_getptstringfromsecondsdouble(s, string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 11
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine getPTStringFromSecondsDouble
  !>
  !! @brief Return a PT String corresponding to arbitrary number of minutes.
  !!
  !! getPTStringFromMinutes() translates minutes values to ISO 8601 compliant timedelta string. 
  !! Conversion of m >= 1440 and  m <= -1440 not supported.
  !!
  !! @param[in]  m
  !!         An int64_t value to be translated.
  !!
  !! @param[out] string
  !!         Translated string is written here.
  !!
  subroutine getPTStringFromMinutes(m, string, errno) !OK-TESTED.
    integer(c_int64_t) :: m
    character(len=*) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    string(:) = ''
    if (present(errno)) errno = 0
    dummy_ptr = my_getptstringfromminutes(m, string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 11      
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine getPTStringFromMinutes
  !>
  !! @brief Return a PT String corresponding to arbitrary number of Hours.
  !!
  !! getPTStringFromHours() translates hour values to ISO 8601 compliant timedelta string. 
  !! Conversion of h >= 24 and  ms <= -24 not supported.
  !!
  !! @param[in]  h
  !!         An int64_t value to be translated.
  !!
  !! @param[out]  string
  !!         Translated string is written here.
  !!
  subroutine getPTStringFromHours(h, string, errno) !OK-TESTED.
    integer(c_int64_t)                   :: h
    character(len=*) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    string(:) = ''
    if (present(errno)) errno = 0
    dummy_ptr = my_getptstringfromhours(h, string)
    if (.not. c_associated(dummy_ptr)) then
      if (present(errno)) errno = 5 * 100 + 11      
      string = '<null>'
    else
      char_loop: do i = 1 , len(string)
        if (string(i:i) == c_null_char) exit char_loop
      end do char_loop
      string(i:len(string)) = ' '
    endif
  end subroutine getPTStringFromHours

  !>
  !! @brief Convert time delta to "Julian calendar delta".
  !!
  SUBROUTINE timeDeltaToJulianDelta(td,dt,jd)
    type(c_ptr) :: dummy_ptr
    TYPE(timedelta), TARGET, INTENT(in) :: td
    TYPE(datetime),  TARGET, INTENT(in) :: dt
    TYPE(juliandelta),  TARGET, INTENT(out) :: jd
    dummy_ptr = my_timedeltatojuliandelta(c_loc(td),c_loc(dt),c_loc(jd))
  END SUBROUTINE timeDeltaToJulianDelta

  !>
  !! @brief division by seconds.
  !!
  !! @param[in]  dividend
  !!         A pointer to type timedelta
  !!
  !! @param[in]  divisor
  !!         A pointer to type timedelta
  !!
  !! @param[out]  quotient
  !!         A pointer to type divisionquotienttimespan
  !!
  subroutine divideTimeDeltaInSeconds(dividend, divisor, quotient)
    type(timedelta), target, intent(in) :: dividend
    type(timedelta), target, intent(in) :: divisor
    type(divisionquotienttimespan), target, intent(out) :: quotient
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_dividetimedeltainseconds(c_loc(dividend), c_loc(divisor), c_loc(quotient))
  end subroutine divideTimeDeltaInSeconds
  !>
  !! @brief division of two differences in datetimes.
  !! 
  !! @param  dt1_dividend, dt2_dividend, dt1_divisor, dt2_divisor
  !!         Reference date (a pointer to struct _datetime).
  !! 
  !! @param  intvlsec
  !!         Interval given in seconds.
  !! 
  !! @return result of division. NULL indicates error.
  subroutine divideTwoDatetimeDiffsInSeconds(dt1_dividend,dt2_dividend, &
      &                                      dt1_divisor, dt2_divisor,  &
      &                                      denominator, quotient)
    type(datetime), target, intent(in) :: dt1_dividend
    type(datetime), target, intent(in) :: dt2_dividend
    type(datetime), target, intent(in) :: dt1_divisor
    type(datetime), target, intent(in) :: dt2_divisor
    integer(c_int64_t), target, intent(out) :: denominator
    type(divisionquotienttimespan), target, intent(out) :: quotient
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_dividetwodatetimediffsinseconds(c_loc(dt1_dividend), c_loc(dt2_dividend),  &
        &                                                c_loc(dt1_divisor), c_loc(dt2_divisor),  &
        &                                                c_loc(denominator), c_loc(quotient))
  end subroutine divideTwoDatetimeDiffsInSeconds
 
  !>
  !! @brief division of an datetime interval by seconds.
  !!
  !! the datetime interval is calculated by dt1-dt2.
  !!
  !! @param[in]  dt1
  !!         A pointer to type datetime
  !!
  !! @param[in]  dt2
  !!         A pointer to type datetime
  !!
  !! @param[in]  divisor
  !!         A pointer to type timedelta
  !!
  !! @param[out]  quotient
  !!         A pointer to type divisionquotienttimespan
  !!
  subroutine divideDatetimeDifferenceInSeconds(dt1, dt2, divisor, quotient)
    type(datetime), target, intent(in) :: dt1
    type(datetime), target, intent(in) :: dt2
    type(timedelta), target, intent(in) :: divisor
    type(divisionquotienttimespan), target, intent(out) :: quotient
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_dividedatetimedifferenceinseconds(c_loc(dt1), c_loc(dt2), c_loc(divisor), c_loc(quotient))
  end subroutine divideDatetimeDifferenceInSeconds
  !
end module mtime_timedelta
!>
!! @brief Definition of the basic event type and its methods.
!!
!! @details
!!
!___________________________________________________________________________________________________________
module mtime_events
  !
  use, intrinsic :: iso_c_binding, only: c_int64_t, c_char, c_null_char, c_bool, c_ptr, &
                                       & c_loc, c_f_pointer, c_associated, c_null_ptr
  !
  use mtime_datetime
  use mtime_timedelta
  !
  implicit none
  !
  private
  !
  public :: max_eventname_str_len
  public :: event
  public :: newEvent
  public :: deallocateEvent
  public :: eventToString
  public :: isCurrentEventActive
  public :: isEventNextInNextDay
  public :: isEventNextInNextMonth
  public :: isEventNextInNextYear
  public :: getTriggerNextEventAtDateTime
  public :: getTriggeredPreviousEventAtDateTime
  public :: getEventId
  public :: getEventName
  public :: getEventReferenceDateTime
  public :: getEventFirstDateTime
  public :: getEventLastDateTime
  public :: getEventInterval
  public :: getNextEventIsFirst
  public :: getEventisFirstInDay
  public :: getEventisFirstInMonth
  public :: getEventisFirstInYear
  public :: getEventisLastInDay
  public :: getEventisLastInMonth
  public :: getEventisLastInYear
  !
  !> provides a string length for toString 
  integer, parameter :: max_eventname_str_len = 132
  !
  type, bind(c) :: event
    type(c_ptr) :: nextEventInGroup
    integer(c_int64_t) :: eventId
    type(c_ptr) :: eventName
    type(c_ptr) :: eventsLastEvaluationDateTime
    type(c_ptr) :: eventReferenceDatetime
    type(c_ptr) :: eventFirstDatetime
    type(c_ptr) :: eventLastDatetime
    type(c_ptr) :: eventInterval
    logical(c_bool) :: triggerCurrentEvent
    logical(c_bool) :: nextEventIsFirst
    logical(c_bool) :: eventisFirstInDay
    logical(c_bool) :: eventisFirstInMonth
    logical(c_bool) :: eventisFirstInYear
    logical(c_bool) :: eventisLastInDay
    logical(c_bool) :: eventisLastInMonth
    logical(c_bool) :: eventisLastInYear
    type(c_ptr) :: triggerNextEventDateTime
    type(c_ptr) :: triggeredPreviousEventDateTime
  end type event
  !
  interface newEvent
    module procedure newEventWithString
    module procedure newEventWithDataTypes
  end interface newEvent
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  interface
    !
    function my_newevent(name, referenceDate, firstdate, lastDate, interval, offset) result(c_pointer) &
         &              bind(c, name='newEvent')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: name
      character(c_char), dimension(*) :: referenceDate
      character(c_char), dimension(*) :: firstDate
      character(c_char), dimension(*) :: lastDate
      character(c_char), dimension(*) :: interval
      character(c_char), dimension(*) :: offset
    end function my_newevent
    !
    function my_neweventwithdatatypes(name, referenceDate, firstdate, lastDate, interval, offset) &
         &                           result(c_pointer) bind(c, name='newEventWithDataType')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: name
      type(c_ptr), value :: referenceDate
      type(c_ptr), value :: firstDate
      type(c_ptr), value :: lastDate
      type(c_ptr), value :: interval
      type(c_ptr), value :: offset
    end function my_neweventwithdatatypes
    !
    subroutine my_deallocateevent(ev) bind(c,name='deallocateEvent')
      import :: c_ptr
      type(c_ptr), value :: ev
    end subroutine my_deallocateevent
    !
    function my_eventtostring(my_event, string) result(string_ptr) bind(c, name='eventToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_event
      character(c_char), dimension(*) :: string
    end function my_eventtostring
    !
    function my_isCurrentEventActive(my_event, my_datetime, plus_slack, minus_slack) &
             &                      result(ret) bind(c, name='isCurrentEventActive')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_datetime
      type(c_ptr), value :: plus_slack
      type(c_ptr), value :: minus_slack
      logical(c_bool) :: ret
    end function my_isCurrentEventActive
    !
    function my_iseventnextinnextday(my_event) result(ret) bind(c, name='iseventNextInNextDay')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_iseventnextinnextday
    !
    function my_iseventnextinnextmonth(my_event) result(ret) bind(c, name='iseventNextInNextMonth')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_iseventnextinnextmonth
    !
    function my_iseventnextinnextyear(my_event) result(ret) bind(c, name='iseventNextInNextYear')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_iseventnextinnextyear
    !
    function my_gettriggernexteventatdatetime(my_event,my_currentdatetime,my_datetime) result(c_pointer) &
    & bind(c, name='getTriggerNextEventAtDateTime')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_currentdatetime
      type(c_ptr), value :: my_datetime
    end function my_gettriggernexteventatdatetime
    !
    function my_gettriggeredpreviouseventatdatetime(my_event,my_datetime) result(c_pointer) &
    & bind(c, name='getTriggeredPreviousEventAtDateTime')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_datetime
    end function my_gettriggeredpreviouseventatdatetime
    !
    function my_geteventname(my_event, string) result(c_pointer) bind(c, name='getEventName')
      import :: c_ptr, c_char
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: my_event
      character(c_char), dimension(*) :: string
    end function my_geteventname
    !
    function my_getnexteventisfirst(my_event) result(ret) bind(c, name='getNextEventIsFirst')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_getnexteventisfirst
    !
    function my_geteventisfirstinday(my_event) result(ret) bind(c, name='getEventisFirstInDay')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventisfirstinday
    !
    function my_geteventisfirstinmonth(my_event) result(ret) bind(c, name='getEventisFirstInMonth')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventisfirstinmonth
    !
    function my_geteventisfirstinyear(my_event) result(ret) bind(c, name='getEventisFirstInYear')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventisfirstinyear
    !
    function my_geteventislastinday(my_event) result(ret) bind(c, name='getEventisLastInDay')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventislastinday
    !
    function my_geteventislastinmonth(my_event) result(ret) bind(c, name='getEventisLastInMonth')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventislastinmonth
    !
    function my_geteventislastinyear(my_event) result(ret) bind(c, name='getEventisLastInYear')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventislastinyear
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !
  !! @param[out]       errno       optional, error message     
  function newEventWithString(name, referenceDate, firstdate, lastDate, interval, offset, errno) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    character(len=*), intent(in)           :: name
    character(len=*), intent(in)           :: referenceDate
    character(len=*), intent(in)           :: firstDate
    character(len=*), intent(in)           :: lastDate
    character(len=*), intent(in)           :: interval
    character(len=*), intent(in), optional :: offset
    character(len=32)                      :: zeroOffset = "PT00S" !Optional offset's proxy string.
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    if(present(offset)) then !offset is not null.
         c_pointer = my_newevent(trim(adjustl(name))//c_null_char,          &
         &                  trim(adjustl(referenceDate))//c_null_char,      &
         &                  trim(adjustl(firstDate))//c_null_char,          &
         &                  trim(adjustl(lastDate))//c_null_char,           &
         &                  trim(adjustl(interval))//c_null_char,           &
         &                  trim(adjustl(offset))//c_null_char)
    else                     !offset is null. Pass a 0 offset. 
         c_pointer = my_newevent(trim(adjustl(name))//c_null_char,    &
         &                  trim(adjustl(referenceDate))//c_null_char,      &
         &                  trim(adjustl(firstDate))//c_null_char,          &
         &                  trim(adjustl(lastDate))//c_null_char,           &
         &                  trim(adjustl(interval))//c_null_char,           &
         &                  trim(adjustl(zeroOffset))//c_null_char)
    end if
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 6 * 100 + 1
    call c_f_pointer(c_pointer, ret_event)
  end function newEventWithString
  !
  !! @param[out]       errno       optional, error message     
  function newEventWithDataTypes(name, referenceDate, firstdate, lastDate, interval, offset, errno) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    character(len=*), intent(in) :: name
    type(datetime), pointer :: referenceDate
    type(datetime), pointer :: firstDate
    type(datetime), pointer :: lastDate
    type(timedelta), pointer :: interval
    type(timedelta), pointer, optional :: offset
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    if (present(offset)) then
    c_pointer = my_neweventwithdatatypes(trim(adjustl(name))//c_null_char, &
         &                  c_loc(referenceDate), c_loc(firstDate), c_loc(lastDate), c_loc(interval), c_loc(offset))
    else
    c_pointer = my_neweventwithdatatypes(trim(adjustl(name))//c_null_char, &
         &                  c_loc(referenceDate), c_loc(firstDate), c_loc(lastDate), c_loc(interval), c_null_ptr)
    end if
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 6 * 100 + 1
    call c_f_pointer(c_pointer, ret_event)
  end function newEventWithDataTypes
  !>
  !! @brief Destructor of Event.
  !!
  !! @param  my_event
  !!         A pointer to type event. my_event is deallocated.
  !!
  !! WARNING: If my_event was added to a group, this should never be called; 
  !! use removeEventFromEventGroup instead. 
  subroutine deallocateEvent(my_event) 
    type(event), pointer :: my_event
    call my_deallocateevent(c_loc(my_event))
    my_event => null()
  end subroutine deallocateEvent
  !>
  !! @brief Get Event as a string.
  !!
  !! @param[in]  my_event
  !!         A pointer to type event. The event to be converted to string.
  !!
  !! @param[out]  string
  !!         String where event is to be written.
  !!
  !! @param[out]       errno       optional, error message     
  subroutine eventToString(my_event, string, errno) !TODO:C code still incomplete.
    type(event), pointer :: my_event
    character(len=max_eventname_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_eventtostring(c_loc(my_event), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 3
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine eventToString
  !>
  !! @brief Check if this event is active by comparing event's trigger time with current_dt.
  !!
  !!        The current_dt must lie in event's trigger time 
  !!
  !!        (subject to optional specified slack: [Trigger_time -
  !!        minus_slack, Trigger_time + plus_slack]. Slacks can be
  !!        NULL. Always inclusive.)
  !!
  !!        The lib has no built-in clock but relies on
  !!        isCurrentEventActive(.) being called (polled) from the
  !!        application at fixed intervals.
  !!
  !! @param  my_event
  !!         A pointer to type event. This is the event being tested.
  !!
  !! @param  my_datetime
  !!         A pointer to type datetime. This is the 'current' datetime of the system.
  !!
  !! @param  plus_slack
  !!         A pointer to type timedelta. Events are triggered between [actual_trigger_time, actual_trigger_time + plus_slack]. 
  !!         Sign MUST be '+'
  !!
  !! @param  minus_slack
  !!         A pointer to type timedelta. Events are triggered between [actual_trigger_time - minus_slack, actual_trigger_time]. 
  !!         Sign MUST be '+'
  !! 
  !! @return ret
  !!        true/false indicating if the event is active.
  function isCurrentEventActive(my_event, my_datetime, plus_slack, minus_slack) result(ret)
    type(event), pointer :: my_event
    type(datetime), target :: my_datetime
    type(timedelta), pointer, optional :: plus_slack
    type(timedelta), pointer, optional :: minus_slack
    logical(c_bool) :: ret
    if (present(plus_slack) .and. present(minus_slack)) then 
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                   c_loc(plus_slack),  c_loc(minus_slack))

    else if (present(plus_slack) .and. .not.present(minus_slack)) then
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                           c_loc(plus_slack),  c_null_ptr)

    else if (.not.present(plus_slack) .and. present(minus_slack)) then
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                          c_null_ptr,  c_loc(minus_slack))
    
    else
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                                   c_null_ptr, c_null_ptr)
    
    end if
  end function isCurrentEventActive
  !>
  !! @brief Checks, if next event is on a new day
  !!
  !! @param[in] my_event
  !!        A pointer to a type event
  !!
  !! @returns ret
  !!        Logical: true if next event is on new day 
  function isEventNextInNextDay(my_event) result(ret)
    type(event), pointer :: my_event
    logical(c_bool) :: ret
    ret = my_iseventnextinnextday(c_loc(my_event))
  end function isEventNextInNextDay
  !>
  !! @brief Checks, if next event is in a new month
  !!
  !! @param[in] my_event
  !!        A pointer to a type event
  !!
  !! @returns ret
  !!        Logical: true if next event is in a new month
  function iseventNextInNextMonth(my_event) result(ret)
    type(event), pointer :: my_event
    logical(c_bool) :: ret
    ret = my_iseventnextinnextmonth(c_loc(my_event))
  end function iseventNextInNextMonth
  !>
  !! @brief Checks, if next event is in a new year  
  !!
  !! @param[in] my_event
  !!        A pointer to a type event
  !!
  !! @returns ret
  !!        Logical: true if next event is in a new year
  function iseventNextInNextYear(my_event) result(ret)
    type(event), pointer :: my_event
    logical(c_bool) :: ret
    ret = my_iseventnextinnextyear(c_loc(my_event))
  end function iseventNextInNextYear
  !>
  !! @brief Get the Datetime when 'this' event will be triggered next.
  !!
  !! WARNING: The value returned is with-respect-to current_dt and not
  !!          a true copy of triggerNextEventDateTime in the event
  !!          data structure.
  !!
  !! @param[in] my_event
  !!         A pointer to type event. This is the event being queried.
  !!
  !! @param[in]  my_currentdatetime
  !!         A pointer to type datetime. The next trigger datetime is copied here.
  !!
  !! @param[out] my_datetime
  !!         A pointer to type datetime with next-trigger datetime.
  !!
  !! @param[out]       errno       optional, error message     
  subroutine getTriggerNextEventAtDateTime(my_event,my_currentdatetime,my_datetime, errno)
    type(event), target, intent(in) :: my_event
    type(datetime), target, intent(in) :: my_currentdatetime
    type(datetime), target :: my_datetime
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_gettriggernexteventatdatetime(c_loc(my_event),c_loc(my_currentdatetime),c_loc(my_datetime))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 8
  end subroutine getTriggerNextEventAtDateTime
  !>
  !! @brief Get the Datetime when 'this' event will be triggered last.
  !!
  !! NOTE: If the event was never tiggered, default value of
  !!       0-01-01T00:00:00.000 is returned.
  !!
  !! @param[in] my_event
  !!         A pointer to type event. This is the event being queried.
  !!
  !! @param[out] my_datetime
  !!         A pointer to type datetime with last-trigger datetime.
  !!
  !! @param[out]       errno       optional, error message     
  subroutine getTriggeredPreviousEventAtDateTime(my_event,my_datetime,errno)
    type(event), target, intent(in) :: my_event
    type(datetime), target :: my_datetime
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_gettriggeredpreviouseventatdatetime(c_loc(my_event),c_loc(my_datetime))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 9
  end subroutine getTriggeredPreviousEventAtDateTime
  !>
  !! @brief get the event id
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret_evtid
  !!        the event id
  !!
  function getEventId(my_event) result(ret_evtid) !OK-TESTED.
    integer(c_int64_t) :: ret_evtid
    type(event), pointer :: my_event
    ret_evtid = my_event%eventId
  end function getEventId
  !>
  !! @brief get the event name
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @param[out]       string      the name of the event
  !!
  !! @param[out]       errno       optional, error message     
  subroutine getEventName(my_event, string, errno) !OK-TESTED.
    type(event), pointer :: my_event
    character(len=max_eventname_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional :: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_geteventname(c_loc(my_event), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 11
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getEventName
  !>
  !! @brief get the event reference date
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret_referenceDateTime
  !!        A pointer of type datetime. The event's reference date.
  !!
  function getEventReferenceDateTime(my_event) result(ret_referenceDateTime) !OK-TESTED.
    type(datetime), pointer :: ret_referenceDateTime
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventReferenceDatetime
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_referenceDateTime)
    else
      ret_referenceDateTime => null()
    endif
  end function getEventReferenceDateTime
  !>
  !! @brief get the event first date
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret_eventFirstDateTime
  !!        A pointer of type datetime. The event's first date.
  !!
  function getEventFirstDateTime(my_event) result(ret_eventFirstDateTime) !OK-TESTED.
    type(datetime), pointer :: ret_eventFirstDateTime
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventFirstDateTime
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_eventFirstDateTime)
    else
      ret_eventFirstDateTime => null()
    endif
  end function getEventFirstDateTime
  !>
  !! @brief get the event last date
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret_eventLastDateTime
  !!        A pointer of type datetime. The event's last date.
  !!
  function getEventLastDateTime(my_event) result(ret_eventLastDateTime) !OK-TESTED.
    type(datetime), pointer :: ret_eventLastDateTime
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventLastDateTime
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_eventLastDateTime)
    else
      ret_eventLastDateTime => null()
    endif
  end function getEventLastDateTime
  !>
  !! @brief get the event interval
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret_eventInterval
  !!        A pointer of type timedelta. The event's last date.
  !!
  function getEventInterval(my_event) result(ret_eventInterval) !OK-TESTED.
    type(timedelta), pointer :: ret_eventInterval
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventInterval
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_eventInterval)
    else
      ret_eventInterval => null()
    endif
  end function getEventInterval
  !>
  !! @brief Check if event is first
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is first
  !!
  function getNextEventIsFirst(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_getnexteventisfirst(c_loc(my_event))
  end function getNextEventIsFirst
  !>
  !! @brief Check if event is first in day
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is first in day
  !!
  function getEventisFirstInDay(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventisfirstinday(c_loc(my_event))
  end function getEventisFirstInDay
  !>
  !! @brief Check if event is first in month
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is first in month
  !!
  function getEventisFirstInMonth(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventisfirstinmonth(c_loc(my_event))
  end function getEventisFirstInMonth
  !>
  !! @brief Check if event is first in year
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is first in year
  !!
  function getEventisFirstInYear(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventisfirstinyear(c_loc(my_event))
  end function getEventisFirstInYear
  !>
  !! @brief Check if event is last in day
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is last in day
  !!

  function getEventisLastInDay(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventislastinday(c_loc(my_event))
  end function getEventisLastInDay
  !>
  !! @brief Check if event is last in month
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is last in month
  !!  
  function getEventisLastInMonth(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventislastinmonth(c_loc(my_event))
  end function getEventisLastInMonth
  !>
  !! @brief Check if event is last in year
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret
  !!        Logical: true if event is last in year
  !!
  function getEventisLastInYear(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventislastinyear(c_loc(my_event))
  end function getEventisLastInYear
  !
end module mtime_events
!>
!! @brief Event-groups which contains a list of events.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_eventgroups
  !
  use, intrinsic :: iso_c_binding, only: c_int64_t, c_ptr, c_char, c_null_char, c_bool, &
                                       & c_loc, c_f_pointer, c_associated
  !
  use mtime_events
  !
  implicit none
  !
  private
  !
  public :: max_groupname_str_len
  public :: eventgroup
  public :: newEventGroup
  public :: deallocateEventGroup
  public :: addEventToEventGroup
  public :: removeEventFromEventGroup
  public :: getEventGroupId
  public :: getEventGroupName
  public :: getFirstEventFromEventGroup
  public :: getNextEventFromEventGroup
  !
  !> provides a string length for toString 
  integer, parameter :: max_groupname_str_len = 132
  !
  type, bind(c) :: eventgroup
    integer(c_int64_t) :: eventGroupId
    type(c_ptr) :: eventGroupName
    type(c_ptr) :: firstEventInGroup
  end type eventgroup
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  interface
    !
    function my_neweventgroup(name) result(c_pointer) bind(c, name='newEventGroup')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: name
    end function my_neweventgroup
    !
    subroutine my_deallocateeventgroup(evgrp) bind(c,name='deallocateEventGroup')
      import :: c_ptr
      type(c_ptr), value :: evgrp
    end subroutine my_deallocateeventgroup
    !
    function my_addeventtoeventgroup(my_event, my_eventgroup) result(ret) bind(c, name='addNewEventToEventGroup')
      import :: c_bool, c_ptr
      logical(c_bool) :: ret
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_eventgroup
    end function my_addeventtoeventgroup
    !
    function my_removeeventfromeventgroup(evname, evgrp) result(ret) bind(c, name='removeEventFromEventGroup')
      import :: c_bool, c_char, c_ptr
      logical(c_bool) :: ret
      character(c_char), dimension(*) :: evname
      type(c_ptr), value :: evgrp
    end function my_removeeventfromeventgroup
    !
    function my_geteventgroupname(my_eventgroup, string) result(string_ptr) bind(c, name='getEventGroupName')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_eventgroup
      character(c_char), dimension(*) :: string
    end function my_geteventgroupname
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS  
  !
contains
  !>
  !! @brief Construct new event-Group using a string.
  !!
  !! @param[in]  name
  !!         This string contains the name of event group.
  !!
  !! @param[out]       errno       optional, error message
  !!
  !! @return ret_eventgroup
  !!         A pointer to an initialized event-Group. 
  !!
  function newEventGroup(name, errno) result(ret_eventgroup) !OK-TESTED.
    type(eventgroup), pointer :: ret_eventgroup 
    character(len=*), intent(in) :: name
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_neweventgroup(trim(adjustl(name))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 7 * 100 + 1
    call c_f_pointer(c_pointer, ret_eventgroup)
  end function newEventGroup
  !>
  !! @brief Destructor of EventGroup.
  !!
  !! @param[in] my_eventgroup
  !!         A pointer to type eventGroup. my_eventgroup is deallocated.
  !!
  subroutine deallocateEventGroup(my_eventgroup) !OK-TESTED.
    type(eventgroup), pointer :: my_eventgroup
    call my_deallocateeventgroup(c_loc(my_eventgroup))
    my_eventgroup => null()
  end subroutine deallocateEventGroup
  !>
  !! @brief Add new event to an eventgroup.
  !!
  !! @param  my_event
  !!         A pointer to type event. The event to be added.
  !!
  !! @param  my_eventgroup
  !!         A pointer to type eventgroup. The eventgroup where the event is added.
  !!
  !! @return ret
  !!         true/false indicating success or failure of addition.
  function addEventToEventGroup(my_event, my_eventgroup) result(ret) !OK-TESTED.
    logical :: ret
    type(event), pointer :: my_event
    type(eventgroup), pointer :: my_eventgroup
    ret = my_addeventtoeventgroup(c_loc(my_event), c_loc(my_eventgroup))
  end function addEventToEventGroup
  !>
  !! @brief Remove event from eventgroup. CRITICAL: Also, deallocate the event.
  !!
  !! @param  my_name
  !!         The name of event to be removed.
  !!
  !! @param  my_eventgroup
  !!         A pointer to  type eventgroup. The eventgroup to which this event belongs.
  !!
  !! @return ret
  !!         true/false indicating success or failure of removal.
  function removeEventfromEventGroup(my_name, my_eventgroup) result(ret) !OK-TESTED.
    logical :: ret
    character(len=*), intent(in) :: my_name
    type(eventgroup), pointer :: my_eventgroup
    ret = my_removeeventfromeventgroup(trim(adjustl(my_name))//c_null_char, c_loc(my_eventgroup))
  end function removeEventFromEventGroup
  !>
  !! @brief Get event group id
  !!
  !! @param[in]  my_eventgroup
  !!         A pointer to  type eventgroup.
  !!
  !! @return ret_grpid
  !!         The event group id
  function getEventGroupId(my_eventgroup) result(ret_grpid) !OK-TESTED.
    integer(c_int64_t) :: ret_grpid
    type(eventgroup), pointer :: my_eventgroup
    ret_grpid = my_eventgroup%eventGroupId
  end function getEventGroupId
  !>
  !! @brief get the event group name
  !!
  !! @param[in] my_eventgroup
  !!        A pointer of type event.
  !!
  !! @param[out]       string      the name of the event group
  !!
  !! @param[out]       errno       optional, error message     
  subroutine getEventGroupName(my_eventgroup, string, errno)  !TESTED-OK.
    type(eventgroup), pointer :: my_eventgroup
    character(len=max_groupname_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_geteventgroupname(c_loc(my_eventgroup), string)
    if ((.not. c_associated(dummy_ptr)) .and. present(errno)) errno = 7 * 100 + 6
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getEventGroupName
  !>
  !! @brief get the first event in event group
  !!
  !! @param[in] my_eventgroup
  !!        A pointer of type eventgroup.
  !!
  !! @returns ret_event
  !!        A pointer of type event. The first event in eventgroup
  function getFirstEventFromEventGroup(my_eventgroup) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    type(eventgroup), pointer :: my_eventgroup
    type(c_ptr) :: c_pointer
    c_pointer = my_eventgroup%firstEventInGroup
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_event)
    else
      ret_event => null()
    endif
  end function getFirstEventFromEventGroup
  !>
  !! @brief get the next event in an event group an event belongs to
  !!
  !! @param[in] my_event
  !!        A pointer of type event.
  !!
  !! @returns ret_event
  !!        A pointer of type event. The next event in eventgroup
  function getNextEventFromEventGroup(my_event) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%nextEventInGroup
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_event)
    else
      ret_event => null()
    endif
  end function getNextEventFromEventGroup
  !
end module mtime_eventgroups
!>
!! @brief Support for handling ISO 8601:2004 repeated time interval strings.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_utilities
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_char, c_null_char
  !
  implicit none
  !
  private
  !
  public :: splitRepetitionString
  public :: getRepetitions
  public :: max_repetition_str_len
  !
  !> provides a string length for the maximum handable repetition string length input   
  integer, parameter :: max_repetition_str_len = 132
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  interface
    !
    function my_getRepetitions(repetitionString) bind(c, name='getRepetitions')
      import :: c_int, c_char
      integer(c_int) :: my_getRepetitions
      character(c_char), dimension(*) :: repetitionString
    end function my_getRepetitions
    !
    subroutine my_splitRepetitionString(recurringTimeInterval, repetitor, start, end, duration) &
         bind(c, name='splitRepetitionString')
      import :: c_char
      character(c_char), dimension(*) :: recurringTimeInterval
      character(c_char), dimension(*) :: repetitor
      character(c_char), dimension(*) :: start
      character(c_char), dimension(*) :: end
      character(c_char), dimension(*) :: duration
    end subroutine my_splitRepetitionString
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS  
  !
contains
  !>
  !! @brief Extract number of repetitions from repetition string part.
  !!
  !! @param[in] repetitionString 
  !!         A repetition string starting with 'R'.
  !!         A string literal can be accepted.
  !!
  !! @return r
  !!         An int representing the number of repetitions.
  !!
  function getRepetitions(repetitionString) RESULT(r)
    integer :: r
    character(len=*), intent(in) :: repetitionString
    r = my_getRepetitions(trim(adjustl(repetitionString))//c_null_char)
  end function getRepetitions    
  !>
  !! @brief Split ISO 8601:2004 repeated time interval strings into base components
  !!
  !! @param[in] recurringTimeInterval
  !!         The string should contain an  ISO 8601:2004 repeated time 
  !!         interval string.
  !!         A string literal can be accepted.
  !! @param[out] repetitor
  !!         Contains the repetitor part of the input string.
  !! @param[out] start
  !!         Contains the start date part of the input string.
  !! @param[out] end
  !!         Contains the end date part of the input string.
  !! @param[out] duration
  !!         Contains the duration part of the input string.
  !! @param[out] lrepetitor
  !!         Logical: true, if repetion is available
  !! @param[out] lstart
  !!         Logical: true, if start is available
  !! @param[out] lend
  !!         Logical: true, if end is available
  !! @param[out] lduration
  !!         Logical: true, if duration is available
  !!
  subroutine splitRepetitionString(recurringTimeInterval, repetitor, start, end, duration, lrepetitor, lstart, lend, lduration)
    character(len=*), intent(in) :: recurringTimeInterval
    character(len=*), intent(out) :: repetitor
    character(len=*), intent(out) :: start
    character(len=*), intent(out) :: end
    character(len=*), intent(out) :: duration
    
    logical, intent(out) :: lrepetitor
    logical, intent(out) :: lstart
    logical, intent(out) :: lend
    logical, intent(out) :: lduration
  
    integer :: i

    lrepetitor = .false.
    lstart = .false.
    lend = .false.
    lduration = .false.
    
    call my_splitRepetitionString(trim(adjustl(recurringTimeInterval))//c_null_char, repetitor, start, end, duration)
    
    if (repetitor(1:1) /= c_null_char) lrepetitor = .true. 
    char_loop1: do i = 1 , len(repetitor)
      if (repetitor(i:i) == c_null_char) exit char_loop1
    end do char_loop1
    repetitor(i:len(repetitor)) = ' '
    
    if (start(1:1) /= c_null_char) lstart = .true. 
    char_loop2: do i = 1 , len(start)
      if (start(i:i) == c_null_char) exit char_loop2
    end do char_loop2
    start(i:len(start)) = ' '
    
    if (end(1:1) /= c_null_char) lend = .true. 
    char_loop3: do i = 1 , len(end)
      if (end(i:i) == c_null_char) exit char_loop3
    end do char_loop3
    end(i:len(end)) = ' '
  
    if (duration(1:1) /= c_null_char) lduration = .true. 
    char_loop4: do i = 1 , len(duration)
      if (duration(i:i) == c_null_char) exit char_loop4
    end do char_loop4
    duration(i:len(duration)) = ' '
  
  end subroutine splitRepetitionString
  !
end module mtime_utilities
!>
!! @brief Register a callback function and print mtime objects by this function
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________

module mtime_print_by_callback
  !
  use mtime_datetime
  use mtime_timedelta  
  !
  implicit none
  !
  private
  !
  public :: register_print_mtime_procedure
  public :: register_finish_mtime_procedure
  public :: print_mtime
  public :: finish_mtime  
  !
  abstract interface
    subroutine registered_print_mtime_procedure(leading_text, message_text)
      character(len=*), intent(in) :: leading_text
      character(len=*), intent(in) :: message_text    
    end subroutine registered_print_mtime_procedure
  end interface
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  procedure(registered_print_mtime_procedure), pointer :: print_message => null()
  !> @endcond DOXYGEN_IGNORE_THIS  
  !
  interface print_mtime
    module procedure print_mtime_datetime
    module procedure print_mtime_timedelta
  end interface print_mtime
  !
  abstract interface
    subroutine registered_finish_mtime_procedure(leading_text, message_text)
      character(len=*), intent(in) :: leading_text
      character(len=*), intent(in) :: message_text    
    end subroutine registered_finish_mtime_procedure
  end interface
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  procedure(registered_finish_mtime_procedure), pointer :: finish_message => null()
  !> @endcond DOXYGEN_IGNORE_THIS  
  !
  interface finish_mtime
    module procedure finish_mtime_datetime
    module procedure finish_mtime_timedelta
  end interface finish_mtime
  ! 
contains
  !
  !>
  !! @brief Adds a callback function (subroutine) for printing a string
  !!        representation of a datetime or timedelta. The callback function
  !!        is required to have two string arguments, a leading text, and
  !!        a message text.
  !!
  !! @param[in] message_procedure a subroutine with two character(len=*) arguments.
  !!
  subroutine register_print_mtime_procedure(message_procedure)
    procedure(registered_print_mtime_procedure) :: message_procedure
    print_message => message_procedure
  end subroutine register_print_mtime_procedure
  !>
  !! @brief Print a datetime with associated text information by the provided callback function.
  !!        Can be used via the generic print_mtime subroutine.
  !!
  !! @param[in] leading_text the leading information, eg. the caller
  !! @param[in] message_text the message text provided
  !! @param[in] this_datetime the mtime datetime to print
  !!
  subroutine print_mtime_datetime(leading_text, message_text, this_datetime)
    character(len=*), intent(in) :: leading_text
    character(len=*), intent(in) :: message_text    
    type(datetime), pointer :: this_datetime
    !
    character(len=max_datetime_str_len)  :: dstring
    !
    call datetimeToString(this_datetime, dstring)
    call print_message(trim(leading_text), trim(message_text)//' '//trim(dstring))    
    ! 
  end subroutine print_mtime_datetime
  !>
  !! @brief Print a timedelta with associated text information by the provided callback function.
  !!        Can be used via the generic print_mtime subroutine.
  !!
  !! @param[in] leading_text the leading information, eg. the caller
  !! @param[in] message_text the message text provided
  !! @param[in] this_timedelta the mtime timedelta to print
  !!        Can be used via the generic print_mtime subroutine.
  !!
  subroutine print_mtime_timedelta(leading_text, message_text, this_timedelta)
    character(len=*), intent(in) :: leading_text
    character(len=*), intent(in) :: message_text    
    type(timedelta), pointer :: this_timedelta
    !
    character(len=max_timedelta_str_len)  :: tdstring
    !
    call timedeltaToString(this_timedelta, tdstring)
    call print_message(trim(leading_text), trim(message_text)//' '//trim(tdstring))    
    ! 
  end subroutine print_mtime_timedelta
  !
  !>
  !! @brief Adds a callback function (subroutine) for printing a
  !!        string representation of a datetime or timedelta and
  !!        finish the program. The callback function is required to
  !!        have two string arguments, a leading text, and a message
  !!        text.
  !!
  !! @param[in] message_procedure a subroutine with two character(len=*) arguments.
  !!
  subroutine register_finish_mtime_procedure(finish_procedure)
    procedure(registered_finish_mtime_procedure) :: finish_procedure
    finish_message => finish_procedure
  end subroutine register_finish_mtime_procedure
  !>
  !! @brief Print a datetime with associated text information by the provided callback function and finish program.
  !!        Can be used via the generic finish_mtime subroutine.
  !!
  !! @param[in] leading_text the leading information, eg. the caller
  !! @param[in] message_text the message text provided
  !! @param[in] this_datetime the mtime datetime to print
  !!
  subroutine finish_mtime_datetime(leading_text, message_text, this_datetime)
    character(len=*), intent(in) :: leading_text
    character(len=*), intent(in) :: message_text    
    type(datetime), pointer :: this_datetime
    !
    character(len=max_datetime_str_len)  :: dstring
    !
    call datetimeToString(this_datetime, dstring)
    call finish_message(trim(leading_text), trim(message_text)//' '//trim(dstring))    
    ! 
  end subroutine finish_mtime_datetime
  !>
  !! @brief Print a timedelta with associated text information by the provided callback function and finish program.
  !!        Can be used via the generic finish_mtime subroutine.
  !!
  !! @param[in] leading_text the leading information, eg. the caller
  !! @param[in] message_text the message text provided
  !! @param[in] this_timedelta the mtime timedelta to print
  !!        Can be used via the generic print_mtime subroutine.
  !!
  subroutine finish_mtime_timedelta(leading_text, message_text, this_timedelta)
    character(len=*), intent(in) :: leading_text
    character(len=*), intent(in) :: message_text    
    type(timedelta), pointer :: this_timedelta
    !
    character(len=max_timedelta_str_len)  :: tdstring
    !
    call timedeltaToString(this_timedelta, tdstring)
    call finish_message(trim(leading_text), trim(message_text)//' '//trim(tdstring))    
    ! 
  end subroutine finish_mtime_timedelta
  !
end module mtime_print_by_callback
!>
!! @}
!>
!! @brief mtime is a compound module making all library components accessible via one module.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int64_t, c_bool, c_ptr, c_char
  !
  use mtime_calendar
  use mtime_juliandelta
  use mtime_julianday
  use mtime_date
  use mtime_time
  use mtime_datetime
  use mtime_timedelta
  use mtime_events
  use mtime_eventgroups
  use mtime_utilities
  use mtime_print_by_callback
  !
  implicit none
  !
  public
  !
#ifdef DOXYGEN_DOCUMENTATION_ONLY
  integer, parameter :: no_of_sec_in_a_day    =    86400  !!< number of seconds per day, defined in C
  integer, parameter :: no_of_sec_in_a_hour   =     3600  !!< number of seconds per hour, defined in C
  integer, parameter :: no_of_sec_in_a_minute =       60  !!< number of seconds per minute, defined in C
  !  
  integer, parameter :: no_of_ms_in_a_day     = 86400000  !!< number of milli-seconds per day, defined in C
  integer, parameter :: no_of_ms_in_half_day  = 43200000  !!< number of milli-seconds per 12 hours, defined in C
  integer, parameter :: no_of_ms_in_a_hour    =  3600000  !!< number of milli-seconds per hour, defined in C
  integer, parameter :: no_of_ms_in_a_minute  =    60000  !!< number of milli-seconds per minute, defined in C
  integer, parameter :: no_of_ms_in_a_second  =     1000  !!< number of milli-seconds per second, defined in C
#endif
  !
  !> @cond DOXYGEN_IGNORE_THIS
  integer(c_int), bind(c,name='NO_OF_SEC_IN_A_DAY') :: no_of_sec_in_a_day       
  integer(c_int), bind(c,name='NO_OF_SEC_IN_A_HOUR') :: no_of_sec_in_a_hour     
  integer(c_int), bind(c,name='NO_OF_SEC_IN_A_MINUTE') :: no_of_sec_in_a_minute 
  !  
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_DAY') :: no_of_ms_in_a_day         
  integer(c_int), bind(c,name='NO_OF_MS_IN_HALF_DAY') :: no_of_ms_in_half_day   
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_HOUR') :: no_of_ms_in_a_hour      
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_MINUTE') :: no_of_ms_in_a_minute  
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_SECOND') :: no_of_ms_in_a_second 
  !> @endcond DOXYGEN_IGNORE_THIS
  !
  !> provides a string length for the maximum error string length  
  integer, parameter :: max_mtime_error_str_len = 132
  !
  !> @cond DOXYGEN_IGNORE_THIS  
  integer, parameter :: no_error                                        = 0,          &

       &                calendar_calendartostring                       = 0*100 +  1, &
 
       &                julianday_newjulianday                          = 1*100 +  1, &
       &                julianday_juliandaytostring                     = 1*100 +  3, &
       
       &                date_newdatefromstring                          = 2*100 +  1, &
       &                date_newdatefromraw_yi8                         = 2*100 +  2, &
       &                date_newdatefromraw                             = 2*100 +  3, &
       &                date_newdatefromconstructandcopy                = 2*100 +  4, &
       &                date_replacedate                                = 2*100 +  6, &
       &                date_datetostring                               = 2*100 +  7, &
       &                date_datetoposixstring                          = 2*100 +  8, &
       
       &                time_newtimefromstring                          = 3*100 +  1, &
       &                time_newtimefromraw                             = 3*100 +  2, &
       &                time_newtimefromconstructandcopy                = 3*100 +  3, &
       &                time_replacetime                                = 3*100 +  5, &
       &                time_timetostring                               = 3*100 +  6, &
       &                time_timetoposixstring                          = 3*100 +  7, &
       
       &                datetime_newdatetimefromstring                  = 4*100 +  1, &
       &                datetime_newdatetimefromraw_yi8                 = 4*100 +  2, &
       &                datetime_newdatetimefromraw                     = 4*100 +  3, &
       &                datetime_newdatetimefromconstructandcopy        = 4*100 +  4, &
       &                datetime_datetimetostring                       = 4*100 +  6, &
       &                datetime_datetimetoposixstring                  = 4*100 +  7, &
       &                datetime_getnoofdaysinmonthdatetime             = 4*100 + 15, &
       &                datetime_getnoofdaysinyeardatetime              = 4*100 + 16, &
       &                datetime_getdayofyearfromdatetime               = 4*100 + 17, &
       &                datetime_getnoofsecondselapsedinmonthdatetime   = 4*100 + 18, &
       &                datetime_getnoofsecondselapsedindaydatetime     = 4*100 + 19, &
       &                datetime_getjuliandayfromdatetime               = 4*100 + 20, &
       &                datetime_getdatetimefromjulianday               = 4*100 + 21, &  
       
       &                timedelta_newtimedeltafromstring                = 5*100 +  1, &
       &                timedelta_newtimedeltafromraw                   = 5*100 +  2, &
       &                timedelta_newtimedeltafromraw_yi8               = 5*100 +  3, &
       &                timedelta_newtimedeltafromconstructandcopy      = 5*100 +  4, &
       &                timedelta_gettotalmillisecondstimedelta         = 5*100 +  8, &
       &                timedelta_gettotalsecondstimedelta              = 5*100 +  9, &
       &                timedelta_timedeltatostring                     = 5*100 + 10, &
       &                timedelta_getptstring                           = 5*100 + 11, &
       
       &                events_newevent                                 = 6*100 +  1, &
       &                events_eventtostring                            = 6*100 +  3, &
       &                events_gettriggernexteventatdatetime            = 6*100 +  8, &
       &                events_gettriggeredpreviouseventatdatetime      = 6*100 +  9, &
       &                events_geteventname                             = 6*100 + 11, &
       
       &                eventgroups_neweventgroup                       = 7*100 +  1, &
       &                eventgroups_geteventgroupname                   = 7*100 +  6
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !>
  !! @brief returns error emssage associated to error number of mtime
  !!
  !! @param[in]   errno           error message number
  !!
  !! @param[out]  error_message   associated error message         
  subroutine mtime_strerror(errno, error_message)
    integer, intent(in) :: errno
    character(len=*), intent(out) :: error_message

    select case (errno) 
    case (no_error) 
      error_message = 'no error'
      
    case (calendar_calendartostring)
      error_message = 'could not retrieve the string in <calendartostring>'
      
    case (julianday_newjulianday)
      error_message = 'could not allocate julianday in <newjulianday>'
    case (julianday_juliandaytostring)
      error_message = 'could not retrieve the string in <juliandaytostring>'
      
    case (date_newdatefromstring)
      error_message = 'could not allocate date in <newdate>'
    case (date_newdatefromraw_yi8)
      error_message = 'could not allocate date in <newdate>'
    case (date_newdatefromraw)
      error_message = 'could not allocate date in <newdate>'
    case (date_newdatefromconstructandcopy)
      error_message = 'could not allocate date in <newdate>'
    case (date_replacedate)
      error_message = 'replace-date failed in <replacedate>'
    case (date_datetostring)
      error_message = 'could not retrieve the string in <datetostring>'
    case (date_datetoposixstring)
      error_message = 'could not retrieve the string in <datetoposixstring>'

    case (time_newtimefromstring)
      error_message = 'could not allocate time in <newtime>'
    case (time_newtimefromraw)
      error_message = 'could not allocate time in <newtime>'
    case (time_newtimefromconstructandcopy)
      error_message = 'could not allocate time in <newtime>'
    case (time_replacetime)
      error_message = 'replace-time failed in <replacetime>'
    case (time_timetostring)
      error_message = 'could not retrieve the string in <timetostring>'
    case (time_timetoposixstring)
      error_message = 'could not retrieve the string in <timetoposixstring>'

    case (datetime_newdatetimefromstring)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_newdatetimefromraw_yi8)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_newdatetimefromraw)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_newdatetimefromconstructandcopy)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_datetimetostring)
      error_message = 'could not retrieve the string in <timetostring>'
    case (datetime_datetimetoposixstring)
      error_message = 'could not retrieve the string in <timetoposixstring>'
    case (datetime_getnoofdaysinmonthdatetime)
      error_message = 'could not retrieve the no-of-days-in-month-datetime in <getnoofdaysinmonthdatetime>'
    case (datetime_getnoofdaysinyeardatetime)
      error_message = 'could not retrieve the no-of-days-in-year-datetime in <getnoofdaysinyeardatetime>'
    case (datetime_getdayofyearfromdatetime)
      error_message = 'could not calculate the day-of-year-from-date-time in <getdayofyearfromdatetime>'
    case (datetime_getnoofsecondselapsedinmonthdatetime)
      error_message = 'could not calculate the no-of-seconds-elapsed-in-month-datetime in <getnoofsecondselapsedinmonthdatetime>'
    case (datetime_getnoofsecondselapsedindaydatetime)
      error_message = 'could not calculate the no-of-seconds-elapsed-in-day-datetime in <getnoofsecondselapsedindaydatetime>'
    case (datetime_getjuliandayfromdatetime)
      error_message = 'could not retrieve the get-julian-day-from-datetime in <getjuliandayfromdatetime>'
    case (datetime_getdatetimefromjulianday)
      error_message = 'could not retrieve the get-datetime-from-julian-day in <getdatetimefromjulianday>'
      
    case (timedelta_newtimedeltafromstring)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_newtimedeltafromraw)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_newtimedeltafromraw_yi8)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_newtimedeltafromconstructandcopy)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_gettotalmillisecondstimedelta)
      error_message = 'error in calculating total-milli-seconds-timedelta in <gettotalmillisecondstimedelta>'
    case (timedelta_gettotalsecondstimedelta)
      error_message = 'error in calculating total-seconds-in-timedelta in <gettotalmillisecondstimedelta>'
    case (timedelta_timedeltatostring)
      error_message = 'could not retrieve the string in <timedeltatostring>'
    case (timedelta_getptstring)
      error_message = 'could not retrieve the string in <getptstring*>'
      
    case (events_newevent)
      error_message = 'could not allocate event in <newevent>'
    case (events_eventtostring)
      error_message = 'could not retrieve the string in <eventtostring>'
    case (events_gettriggernexteventatdatetime)
      error_message = 'failed to retrieve trigger-next-event-at-datetime in <gettriggernexteventatdatetime>'
    case (events_gettriggeredpreviouseventatdatetime)
      error_message = 'failed to retrieve triggered-previous-event-at-date-time in <gettriggeredpreviouseventatdatetime>'
    case (events_geteventname)
      error_message = 'failed to retrieve event-name in <geteventname>'
      
    case (eventgroups_neweventgroup)
      error_message = 'could not allocate eventgroup in <neweventgroup>'
    case (eventgroups_geteventgroupname)
      error_message = 'could not retrieve the group-name in <geteventgroupname>'
    end select
  end subroutine mtime_strerror
  !
end module mtime
